import logging

from aiohttp import web
from heaserver.service import appproperty, requestproperty
from aiohttp_remotes import XForwardedRelaxed


@web.middleware
async def new_wstl_builder(request: web.Request, handler) -> web.Response:
    wstl_builder_factory = request.app[appproperty.HEA_WSTL_BUILDER_FACTORY]
    request[requestproperty.HEA_WSTL_BUILDER] = wstl_builder_factory()
    response = await handler(request)
    return response


def new_app() -> web.Application:
    """
    Creates and returns an aiohttp Application object. Installs middleware that sets the HEA_WSTL_BUILDER request
    property, assuming that the HEA_WSTL_BUILDER_FACTORY app property has already been set.

    :return: the Application property.
    """
    return web.Application(middlewares=[new_wstl_builder, XForwardedRelaxed().middleware])




from typing import Optional, Mapping, Tuple, Union, Type, TypeVar

from aiohttp.web_request import Request
from heaobject.volume import Volume, FileSystem
from yarl import URL

from .. import client
from ..heaobjectsupport import type_to_resource_url
from ..oidcclaimhdrs import SUB
from heaobject.keychain import Credentials
from heaobject.volume import FileSystemTypeVar


async def get_file_system_and_credentials_from_volume(request: Request, volume_id: Optional[str], file_system_type: Type[FileSystemTypeVar]) -> Tuple[FileSystemTypeVar, Optional[Credentials]]:
    """
    Get the file system and credentials for the given volume.

    :param request: the aiohttp request (required).
    :param volume_id: a volume id string (required).
    :param file_system_type: the type of file system (required).
    :return: a tuple containing a FileSystem object and, if one exists, a Credentials object (or None if one does not).
    :raise ValueError: if no volume with that id exists, no file system exists for the given volume, or the volumes's
    credentials were not found.
    """
    headers = {SUB: request.headers.get(SUB)} if SUB in request.headers else None
    volume, volume_url = await _get_volume(request, volume_id, headers)
    if volume is None:
        raise ValueError(f'No volume with id {volume_id}')
    if volume_url is None:
        raise ValueError(f'Volume {volume_id} has no URL')
    fs_url = await type_to_resource_url(request, FileSystem)
    if fs_url is None:
        raise ValueError('No file system service registered')
    file_system = await client.get(request.app,
                                   URL(fs_url) / 'bytype' / file_system_type.get_type_name() / 'byname' / volume.file_system_name, file_system_type,
                                   headers=headers)
    if file_system is None:
        raise ValueError(f"Volume {volume.id}'s file system {volume.file_system_name} does not exist")
    return file_system, await _get_credentials(request, volume, headers)


async def _get_volume(request: Request, volume_id: Optional[str], headers: Optional[Mapping] = None) -> Tuple[Optional[Volume], Optional[Union[str, URL]]]:
    """
    Gets the volume with the provided id.

    :param request: the HTTP request (required).
    :param volume_id: the id string of a volume.
    :param headers: any headers.
    :return: a two-tuple with either the Volume and its URL, or (None, None).
    :raise ValueError: if there is no volume with the provided volume id or no volume service is registered.
    """
    if volume_id is not None:
        volume_url = await type_to_resource_url(request, Volume)
        if volume_url is None:
            raise ValueError('No Volume service registered')
        volume = await client.get(request.app, URL(volume_url) / volume_id, Volume, headers=headers)
        if volume is None:
            raise ValueError(f'No volume with volume_id={volume_id}')
        return volume, volume_url
    else:
        return None, None


async def _get_credentials(request: Request, volume: Volume, headers: Optional[Mapping] = None) -> Optional[Credentials]:
    """
    Gets a credential specified in the provided volume, or if there is none, a credential with the where attribute set
    to the volume's URL.

    :param request: the HTTP request (required).
    :param volume: the Volume (required).
    :param volume_url: the volume's URL (required).
    :param headers: any headers.
    :return: the Credentials, or None if the volume has no credentials.
    :raise ValueError: if no credentials service is registered or if the volume's credentials were not found.
    """
    if volume.credential_id is not None:
        cred_url = await type_to_resource_url(request, Credentials)
        if cred_url is None:
            raise ValueError('No credentials service registered')
        credential = await client.get(request.app, URL(cred_url) / volume.credential_id, Credentials, headers=headers)
        if credential is not None:
            return credential
        else:
            raise ValueError(f'Credentials {volume.credential_id} not found')
    else:
        return None



"""
Functions for interacting with Amazon Web Services.

This module supports management of AWS accounts, S3 buckets, and objects in S3 buckets. It uses Amazon's boto3 library
behind the scenes.

In order for HEA to access AWS accounts, buckets, and objects, there must be a volume accessible to the user through
the volumes microservice with an AWSFileSystem for its file system. Additionally, credentials must either be stored
in the keychain microservice and associated with the volume through the volume's credential_id attribute,
or stored on the server's file system in a location searched by the AWS boto3 library. Users can only see the
accounts, buckets, and objects to which the provided AWS credentials allow access, and HEA may additionally restrict
the returned objects as documented in the functions below. The purpose of volumes in this case is to supply credentials
to AWS service calls. Support for boto3's built-in file system search for credentials is only provided for testing and
should not be used in a production setting. This module is designed to pass the current user's credentials to AWS3, not
to have application-wide credentials that everyone uses.

The request argument to these functions is expected to have a OIDC_CLAIM_sub header containing the user id for
permissions checking. No results will be returned if this header is not provided or is empty.

In general, there are two flavors of functions for getting accounts, buckets, and objects. The first expects the id
of a volume as described above. The second expects the id of an account, bucket, or bucket and object. The latter
attempts to match the request up to any volumes with an AWSFileSystem that the user has access to for the purpose of
determine what AWS credentials to use. They perform the
same except when the user has access to multiple such volumes, in which case supplying the volume id avoids a search
through the user's volumes.
"""
import asyncio
import logging
import boto3
from botocore.exceptions import ClientError
import os
from aiohttp import web, hdrs
from .. import response, client
from heaobject.root import type_for_name
from heaserver.service.appproperty import HEA_DB
from heaserver.service.heaobjectsupport import type_to_resource_url, new_heaobject_from_type

from .. import response, client
from .servicelib import get_file_system_and_credentials_from_volume
from ..heaobjectsupport import type_to_resource_url
from ..oidcclaimhdrs import SUB
from typing import Any, Optional, List, Dict, Union
from aiohttp.web import Response
from aiohttp.web import Request
from heaobject.volume import AWSFileSystem, Volume
from heaobject.user import NONE_USER
from heaobject.bucket import AWSBucket, Bucket
from heaobject.root import DesktopObjectDict
from yarl import URL
from asyncio import gather
from heaobject.root import Tag

"""
Available functions
AWS object
- get

- change_storage_class            TODO
- copy_object
- delete_bucket_objects
- delete_bucket
- delete_folder
- delete_object
- download_object
- download_archive_object         TODO
- generate_presigned_url
- get_object_meta
- get_object_content
- get_all_buckets
- get all
- opener                          TODO -> return file format -> returning metadata containing list of links following collection + json format
-                                         need to pass back collection - json format with link with content type, so one or more links, most likely
- post_bucket
- post_folder
- post_object
- post_object_archive             TODO
- put_bucket
- put_folder
- put_object
- put_object_archive              TODO
- transfer_object_within_account
- transfer_object_between_account TODO
- rename_object
- update_bucket_policy            TODO

TO DO
- accounts?
"""
MONGODB_BUCKET_COLLECTION = 'buckets'


async def get_account(request: Request, volume_id: str):
    """
    Gets the AWS account associated with the provided volume id.

    Only get since you can't delete or put id information
    currently being accessed. If organizations get included, then the delete, put, and post will be added for name,
    phone, email, ,etc.
    NOTE: maybe get email from the login portion of the application?

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :return: account object with
        account id
        account name
        full name
        phone number
        alternate contact name
        alternate email address
        alternate phone number
        charge account?
    FIXME: a bad volume_id should result in a 400 status code; currently has status code 500.
    """
    aws_object_dict = await _get_account(request, volume_id)
    return await response.get(request, aws_object_dict)


async def get_account_by_id(request: web.Request) -> Optional[DesktopObjectDict]:
    """
    Gets an account by its id. The id is expected to be the request object's match_info mapping, with key 'id'.

    :param request: an aiohttp Request object (required).
    :return: an AWSAccount dict.
    """
    headers = {SUB: request.headers.get(SUB)} if SUB in request.headers else None
    volume_url = await type_to_resource_url(request, Volume)
    if volume_url is None:
        raise ValueError('No Volume service registered')
    get_volumes_url = URL(volume_url) / 'byfilesystemtype' / AWSFileSystem.get_type_name()

    async def get_one(request, volume_id):
        return await _get_account(request, volume_id)

    return next((a for a in await gather(
        *[get_one(request, v.id) async for v in client.get_all(request.app, get_volumes_url, Volume, headers=headers)])
                 if
                 a['account_id'] == request.match_info['id']), None)


async def get_volume_id_for_account_id(request: web.Request) -> Optional[str]:
    """
    Gets the id of the volume associated with an AWS account. The account id is expected to be in the request object's
    match_info mapping, with key 'id'.

    :param request: an aiohttp Request object (required).
    :return: a volume id string, or None if no volume was found associated with the AWS account.
    """
    headers = {SUB: request.headers.get(SUB)} if SUB in request.headers else None
    volume_url = await type_to_resource_url(request, Volume)
    if volume_url is None:
        raise ValueError('No Volume service registered')
    get_volumes_url = URL(volume_url) / 'byfilesystemtype' / AWSFileSystem.get_type_name()

    async def get_one(request, volume_id):
        return volume_id, await _get_account(request, volume_id)

    return next((volume_id for (volume_id, a) in await gather(
        *[get_one(request, v.id) async for v in client.get_all(request.app, get_volumes_url, Volume, headers=headers)])
                 if
                 a['account_id'] == request.match_info['id']), None)


async def get_all_accounts(request: web.Request) -> List[DesktopObjectDict]:
    """
    Gets all AWS accounts for the current user.

    In order for HEA to access an AWS account, there must be a volume accessible to the user through the volumes
    microservice with an AWSFileSystem for its file system, and credentials must either be stored in the keychain
    microservice and associated with the volume, or stored on the server's file system in a location searched by the
    AWS boto3 library.

    :param request: an aiohttp Request object (required).
    :return: a list of AWSAccount objects, or the empty list of the current user has no accounts.
    """
    headers = {SUB: request.headers.get(SUB)} if SUB in request.headers else None
    volume_url = await type_to_resource_url(request, Volume)
    if volume_url is None:
        raise ValueError('No Volume service registered')
    get_volumes_url = URL(volume_url) / 'byfilesystemtype' / AWSFileSystem.get_type_name()

    async def get_one(request, volume_id):
        return await _get_account(request, volume_id)

    return [a for a in await gather(
        *[get_one(request, v.id) async for v in client.get_all(request.app, get_volumes_url, Volume, headers=headers)])]


async def post_account(request: Request, volume_id: str):
    """
    Placeholder for when this may get implemented in the future.
    """
    return response.status_not_found()


async def put_account(request: Request, volume_id: str):
    """
    Placeholder for when this may get implemented in the future.
    """
    return response.status_not_found()


async def delete_account(request: Request, volume_id: str):
    """
    Placeholder for when this may get implemented in the future.
    """
    return response.status_not_found()


def change_storage_class():
    """
    change storage class (Archive, un-archive) (copy and delete old)

    S3 to archive -> https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/glacier.html#Glacier.Client.upload_archive
        save archive id for future access?
        archived gets charged minimum 90 days
        buckets = vault?
        delete bucket
    archive to S3
        create vault? link vault to account as attribute?
        delete vault
    """


async def copy_object(request: Request, volume_id: str, source_path: str, destination_path: str):
    """
    copy/paste (duplicate), throws error if destination exists, this so an overwrite isn't done
    throws another error is source doesn't exist
    https://medium.com/plusteam/move-and-rename-objects-within-an-s3-bucket-using-boto-3-58b164790b78
    https://stackoverflow.com/questions/47468148/how-to-copy-s3-object-from-one-bucket-to-another-using-python-boto3

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param source_path: (str) s3 path of object, includes bucket and key values together
    :param destination_path: (str) s3 path of object, includes bucket and key values together
    """
    # Copy object A as object B
    s3_resource = await _get_resource(request, 's3', volume_id)
    source_bucket_name = source_path.partition("/")[0]
    source_key_name = source_path.partition("/")[2]
    copy_source = {'Bucket': source_bucket_name, 'Key': source_key_name}
    destination_bucket_name = destination_path.partition("/")[0]
    destination_key_name = destination_path.partition("/")[2]
    try:
        s3_client = await _get_client(request, 's3', volume_id)
        s3_client.head_object(Bucket=destination_bucket_name,
                              Key=destination_key_name)  # check if destination object exists, if doesn't throws an exception
        return web.HTTPBadRequest()
    except ClientError as e:
        error_code = e.response['Error']['Code']
        if error_code == '404':  # object doesn't exist
            try:
                s3_client = await _get_client(request, 's3', volume_id)
                s3_client.head_object(Bucket=source_bucket_name,
                                      Key=source_key_name)  # check if source object exists, if not throws an exception
                s3_resource.meta.client.copy(copy_source, destination_path.partition("/")[0],
                                             destination_path.partition("/")[2])
                logging.info(e)
                return web.HTTPCreated()
            except ClientError as e_:
                logging.error(e_)
                return web.HTTPBadRequest()
        else:
            logging.info(e)
            return web.HTTPBadRequest()


async def delete_bucket_objects(request: Request, volume_id: str, bucket_name: str):
    """
    Deletes all objects inside a bucket

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param bucket_name: Bucket to delete
    """
    try:
        s3_resource = await _get_resource(request, 's3', volume_id)
        s3_client = await _get_client(request, 's3', volume_id)
        s3_client.head_bucket(Bucket=bucket_name)
        s3_bucket = s3_resource.Bucket(bucket_name)
        bucket_versioning = s3_resource.BucketVersioning(bucket_name)
        if bucket_versioning.status == 'Enabled':
            del_obj_all_result = s3_bucket.object_versions.delete()
            logging.info(del_obj_all_result)
        else:
            del_obj_all_result = s3_bucket.objects.all().delete()
            logging.info(del_obj_all_result)
        return web.HTTPNoContent()
    except ClientError as e:
        logging.error(e)
        return web.HTTPNotFound()


async def delete_bucket(request: Request):
    """
    Deletes bucket and all contents
    :param request: the aiohttp Request (required).
    """
    volume_id = request.match_info.get("volume_id", None)
    bucket_id = request.match_info.get("id", None)
    if not volume_id:
        return web.HTTPBadRequest(body="volume_id is required")
    if not bucket_id:
        return web.HTTPBadRequest(body="bucket_id is required")

    s3_client = await _get_client(request, 's3', volume_id)
    try:
        s3_client.head_bucket(Bucket=bucket_id)
        delete_bucket_objects(request, volume_id, bucket_id)
        del_bucket_result = s3_client.delete_bucket(Bucket=bucket_id)
        logging.info(del_bucket_result)
        return web.HTTPNoContent()
    except ClientError as e:
        logging.error(e)
        return web.HTTPNotFound()


async def delete_folder(request: Request, volume_id: str, path_name: str):
    """
    Deletes folder and all contents

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param path_name: path to delete folder, split into bucket and folder for function

    https://izziswift.com/amazon-s3-boto-how-to-delete-folder/
    """
    # TODO: bucket.object_versions.filter(Prefix="myprefix/").delete()     add versioning option like in the delete bucket?
    # TODO: throws error if bucket doesn't exist, should this be desired effect?
    bucket_name = path_name.partition("/")[0]
    folder_name = path_name.partition("/")[2]
    s3_client = await _get_client(request, 's3', volume_id)
    try:
        s3_client.head_object(Bucket=bucket_name,
                              Key=(folder_name + '/'))  # check if folder exists, if not throws an exception
        response = s3_client.list_objects_v2(Bucket=bucket_name, Prefix=folder_name + '/', MaxKeys=10000000)
        for object_f in response['Contents']:
            s3_client.delete_object(Bucket=bucket_name, Key=object_f['Key'])
        delete_folder_result = s3_client.delete_object(Bucket=bucket_name, Key=folder_name + '/')
        logging.info(delete_folder_result)
        return web.HTTPNoContent()
    except ClientError as e:
        logging.error(e)
        return web.HTTPNotFound()


async def delete_object(request: Request, volume_id: str, bucket_name: str, path_name: str):
    """
    Deletes a single object, checks if object exists before deleting, throws error if it doesn't

    :param request: the aiohttp Request (required).
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.delete_object
    """
    s3_client = await _get_client(request, 's3', volume_id)
    volume_id = request.match_info.get("volume_id", None)
    bucket_name = request.match_info.get("bucket_name", None)
    path = request.match_info.get("path", None)

    if not volume_id:
        return web.HTTPBadRequest(body="volume_id is required")
    if not bucket_id:
        return web.HTTPBadRequest(body="bucket_name is required")
    if not path:
        return web.HTTPBadRequest(body="path is required")

    try:
        # TODO: when last object is deleted the folder is also deleted. Should this be saved?
        s3_client.head_object(Bucket=bucket_name, Key=path)  # check if object exists, if not throws an exception
        delete_response = s3_client.delete_object(Bucket=bucket_name, Key=path)
        logging.info(delete_response)
        return web.HTTPNoContent()
    except ClientError as e:
        logging.error(e)
        return web.HTTPBadRequest()


async def download_object(request: Request, volume_id: str, object_path: str, save_path: str):
    r"""
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.download_file

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param object_path: path to the object to download
    :param save_path: path of where object is to be saved, note, needs to include the name of the file to save to
        ex: r'C:\Users\...\Desktop\README.md'  not  r'C:\Users\...\Desktop\'
    """
    try:
        s3_resource = await _get_resource(request, 's3', volume_id)
        bucket_name = object_path.partition("/")[0]
        folder_name = object_path.partition("/")[2]
        s3_resource.meta.client.download_file(bucket_name, folder_name, save_path)
    except ClientError as e:
        logging.error(e)


def download_archive_object(length=1):
    """

    """


def get_archive():
    """
    Don't think it is worth it to have a temporary view of data, expensive and very slow
    """


async def account_opener(request: Request, volume_id: str) -> Response:
    """
    Gets choices for opening an account object.

    :param request: the HTTP request. Required. If an Accepts header is provided, MIME types that do not support links
    will be ignored.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: a Response object with status code 300, and a body containing the HEA desktop object and links
    representing possible choices for opening the HEA desktop object; or Not Found.
    """
    result = await _get_account(request, volume_id)
    if result is None:
        return response.status_not_found()
    return await response.get_multiple_choices(request, result)


async def account_opener_by_id(request: web.Request) -> web.Response:
    """
    Gets choices for opening an account object, using the 'id' value in the match_info attribute of the request.

    :param request: the HTTP request, must contain an id value in its match_info attribute. Required. If an Accepts
    header is provided, MIME types that do not support links will be ignored.
    :return: a Response object with status code 300, and a body containing the HEA desktop object and links
    representing possible choices for opening the HEA desktop object; or Not Found.
    """
    result = await get_account_by_id(request)
    if result is None:
        return response.status_not_found()
    return await response.get_multiple_choices(request, result)


async def generate_presigned_url(request: Request, volume_id: str, path_name: str, expiration: int = 3600):
    """Generate a presigned URL to share an S3 object

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param path_name: string
    :param expiration: Time in seconds for the presigned URL to remain valid
    :return: Presigned URL as string. If error, returns None.

    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-presigned-urls.html
    """
    # Generate a presigned URL for the S3 object
    try:
        s3_client = await _get_client(request, 's3', volume_id)
        bucket_name = path_name.partition("/")[0]
        folder_name = path_name.partition("/")[2]
        response = s3_client.generate_presigned_url('get_object', Params={'Bucket': bucket_name, 'Key': folder_name},
                                                    ExpiresIn=expiration)
        logging.info(response)
    except ClientError as e:
        logging.error(e)
        return None
    # The response contains the presigned URL
    return response


async def get_object_meta(request: Request, volume_id: str, path_name: str):
    """
    preview object in object explorer

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param path_name: path to the object to get, includes both bucket and key values
    """
    try:
        s3_client = await _get_client(request, 's3', volume_id)
        bucket_name = path_name.partition("/")[0]
        folder_name = path_name.partition("/")[2]
        resp = s3_client.get_object(Bucket=bucket_name, Key=folder_name)
        logging.info(resp["ResponseMetadata"])
    except ClientError as e:
        logging.error(e)
        return web.HTTPNotFound()
    return response.get(request=request, data=resp["ResponseMetadata"])


async def get_object_content(request: Request):
    """
    preview object in object explorer
    :param request: the aiohttp Request (required).
    """
    volume_id = request.match_info.get("volume_id", None)
    bucket_name = request.match_info.get("bucket_name", None)
    path = request.match_info.get("path", None)

    if not volume_id:
        return web.HTTPBadRequest(body="volume_id is required")
    if not bucket_name:
        return web.HTTPBadRequest(body="bucket_name is required")
    if not path:
        return web.HTTPBadRequest(body="path is required")

    try:
        s3_client = await _get_client(request, 's3', volume_id)
        resp = s3_client.get_object(Bucket=bucket_name, Key=path_name)
        logging.info(resp["ResponseMetadata"])
        content_type = resp['ContentType']
        body = resp['Body']

    except ClientError as e:
        logging.error(e)
        return web.HTTPNotFound()
    # FIXME need to figure out how to pass a SupportsAsyncRead possibly a wrapper around StreamBody
    # return await response.get_streaming(request=request, out=body, content_type=content_type)


async def get_bucket(request: Request) -> web.Response:
    """
    List a single bucket's attributes

    :param request: the aiohttp Request (required).
    :return:  return the single bucket object requested or HTTP Error Response
    """
    volume_id = request.match_info.get('volume_id', None)
    bucket_id = request.match_info.get('id', None)
    bucket_name = request.match_info.get('bucket_name', None)
    s3_client = await _get_client(request, 's3', volume_id)
    s3_resource = await _get_resource(request, 's3', volume_id)

    bucket_result = await _get_bucket(volume_id=volume_id, s3_resource=s3_resource, s3_client=s3_client,
                                      bucket_name=bucket_name, bucket_id=bucket_id, )
    if type(bucket_result) is AWSBucket:
        return await response.get(request=request, data=bucket_result.to_dict())
    return bucket_result


async def get_all_buckets(request: web.Request) -> List[DesktopObjectDict]:
    """
    List available buckets by name

    :param request: the aiohttp Request (required).
    :return: (list) list of available buckets
    """

    try:
        volume_id = request.match_info.get("volume_id", None)
        if not volume_id:
            return web.HTTPBadRequest(body="volume_id is required")
        s3_client = await _get_client(request, 's3', volume_id)
        s3_resource = await _get_resource(request, 's3', volume_id)

        resp = s3_client.list_buckets()
        bucket_dict = {}
        async_bucket_list = []
        for bucket in resp['Buckets']:
            async_bucket_list.append(_get_bucket(volume_id=volume_id, bucket_name=bucket["Name"],
                                                 s3_client=s3_client, s3_resource=s3_resource,
                                                 creation_date=bucket['CreationDate']))

        buck_list = await asyncio.gather(*async_bucket_list)
    except ClientError as e:
        logging.error(e)
        return response.status_bad_request()
    bucket_dict_list = [buck.to_dict() for buck in buck_list if buck is not None]

    return await response.get_all(request, bucket_dict_list)


async def get_all(request: Request, by_dir_level: bool = False):
    """
    List all objects in  a folder or entire bucket.
    :param by_dir_level: (str) used to switch off recursive nature get all object/folder within level
    :param request: the aiohttp Request (required).
    """
    bucket_name = request.match_info.get('bucket_name', None)
    volume_id = request.match_info.get("volume_id", None)
    if not volume_id:
        return web.HTTPBadRequest(body="volume_id is required")
    if not bucket_name:
        return web.HTTPBadRequest(body="bucket_name is required")

    query_keys = {'max_keys': 500, 'page_size': 50}
    if request.rel_url and request.rel_url.query:
        for key in request.rel_url.query:
            query_keys[key] = request.rel_url.query[key]

    try:
        object_list = []
        pag_request_params = {}
        s3_client = await _get_client(request, 's3', volume_id)
        paginator = s3_client.get_paginator('list_objects_v2')

        pag_request_params['Bucket'] = bucket_name
        pag_request_params['PaginationConfig'] = {'MaxItems': query_keys['max_keys'],
                                                  'PageSize': query_keys['page_size']}

        if 'folder_name' in query_keys and query_keys['folder_name']:
            folder_name = query_keys['folder_name']
            pag_request_params['Prefix'] = folder_name if folder_name.endswith('/') else folder_name + '/'
            pag_request_params['Delimiter'] = '/'
        elif by_dir_level:
            pag_request_params['Delimiter'] = '/'

        if 'start_after_key' in query_keys:
            pag_request_params['StartAfter'] = query_keys['start_after_key']

        if 'starting_token' in query_keys:
            pag_request_params['PaginationConfig']['StartingToken'] = query_keys['starting_token']

        pag_iter = paginator.paginate(**pag_request_params)
        # response = s3_client.list_objects_v2(Bucket=bucket_name, MaxKeys=max_keys)

        results = []
        for page in pag_iter:
            object_list = []
            folder_list = []
            next_token = page['NextContinuationToken'] if page['IsTruncated'] else None
            current_token = page['ContinuationToken'] if 'ContinuationToken' in page else None
            content = page.get('Contents', [])
            common_prefixes = page.get('CommonPrefixes', {})
            for pre in common_prefixes:
                folder: str = pre['Prefix']
                level = len(page['Prefix'].split('/'))
                if level == 1:
                    folder_list.append({'type': 'project', 'folder': folder})
                elif level == 2:
                    folder_list.append({'type': 'analysis', 'folder': folder})
                else:
                    folder_list.append({'type': 'general', 'folder': folder})

            start_after_key = content[len(content) - 1]['Key'] if len(content) > 0 else None
            for val in content:
                object_list.append(val)

            results.append({'NextContinuationToken': next_token,
                            'ContinuationToken': current_token,
                            'StartAfter': start_after_key,
                            'Objects': object_list,
                            'Folders': folder_list
                            })
    except ClientError as e:
        logging.error(e)
        return web.HTTPNotFound
    return await response.get(request, results)


async def post_bucket(request: Request):
    """
    Create an S3 bucket in a specified region, checks that it is the first, if already exists errors are thrown
    If a region is not specified, the bucket is created in the S3 default region (us-east-1).

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    """
    volume_id = request.match_info.get('volume_id', None)
    if not volume_id:
        return web.HTTPBadRequest(body="volume_id is required")
    b: AWSBucket = await new_heaobject_from_type(request=request, type_=AWSBucket)
    if not b:
        return web.HTTPBadRequest(body="Post body is not an HEAObject AWSBUCKET")
    if not b.name:
        return web.HTTPBadRequest(body="Bucket name is required")

    s3_client = await _get_client(request, 's3', volume_id)
    try:
        resp = s3_client.head_bucket(Bucket=b.name)  # check if bucket exists, if not throws an exception
        logging.info(resp)
        return web.HTTPBadRequest(body="bucket already exists")
    except ClientError as e:
        try:
            # todo this is a privileged actions need to check if authorized
            error_code = e.response['Error']['Code']

            if error_code == '404':  # bucket doesn't exist
                create_bucket_params = {'Bucket': b.name}
                put_bucket_policy_params = {
                    'BlockPublicAcls': True,
                    'IgnorePublicAcls': True,
                    'BlockPublicPolicy': True,
                    'RestrictPublicBuckets': True
                }
                if b.region:
                    create_bucket_params['CreateBucketConfiguration'] = {'LocationConstraint': b.region}
                if b.locked:
                    create_bucket_params['ObjectLockEnabledForBucket'] = True

                create_bucket_result = s3_client.create_bucket(**create_bucket_params)
                # make private bucket
                s3_client.put_public_access_block(Bucket=b.name,
                                                  PublicAccessBlockConfiguration=put_bucket_policy_params)

                if b.encrypted:
                    try:
                        SSECNF = 'ServerSideEncryptionConfigurationNotFoundError'
                        s3_client.get_bucket_encryption(Bucket=b.name)
                    except ClientError as e:
                        if e.response['Error']['Code'] == SSECNF:
                            config = \
                                {'Rules': [{'ApplyServerSideEncryptionByDefault':
                                                {'SSEAlgorithm': 'AES256'}, 'BucketKeyEnabled': False}]}
                            s3_client.put_bucket_encryption(Bucket=b.name, ServerSideEncryptionConfiguration=config)
                        else:
                            logging.error(e.response['Error']['Code'])
                            raise e
                # todo this is a privileged action need to check if authorized ( may only be performed by bucket owner)

                vresp = await put_bucket_versioning(request=request, volume_id=volume_id,
                                                    bucket_name=b.name, s3_client=s3_client, is_versioned=b.versioned)
                if vresp.status == 400:
                    raise ClientError(vresp.body)
                logging.debug(vresp)

                tag_resp = await put_bucket_tags(request=request, volume_id=volume_id,
                                                 bucket_name=b.name, new_tags=b.tags)
                if tag_resp.status == 400:  # a soft failure, won't cause bucket to be deleted
                    logging.error(tag_resp)
                    return web.HTTPBadRequest(body="Bucket was created but tags failed to be added")

            elif error_code == '403':  # already exists
                logging.error(e)
                return web.HTTPBadRequest(body="bucket exists, no permission to access")
            else:
                logging.error(e)
                return web.HTTPBadRequest()
        except (ClientError, Exception) as e:
            logging.error(e.response)
            try:
                s3_client.head_bucket(Bucket=b.name)
                del_bucket_result = s3_client.delete_bucket(Bucket=b.name)
                logging.info(f"deleted failed bucket {b.name} details: \n{del_bucket_result}")
            except ClientError as e:  # bucket doesn't exist so no clean up needed
                pass
            return web.HTTPBadRequest(body="New bucket was NOT created")
        return web.HTTPCreated()


async def put_bucket_versioning(request: Request, volume_id: str, bucket_name: str,
                                is_versioned: bool, s3_client: Optional[boto3.client] = None):
    """
    Use To change turn on or off bucket versioning settings. Note that if the Object Lock
    is turned on for the bucket you can't change these settings.

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param bucket_name: The bucket name
    :param is_versioned: For toggling on or off the versioning
    :param s3_client: Pass the active client if exists (optional)
    """

    try:
        s3_client = s3_client if s3_client else await _get_client(request, 's3', volume_id)
        vconfig = {
            'MFADelete': 'Disabled',
            'Status': 'Enabled' if is_versioned else 'Suspended',
        }
        vresp = s3_client.put_bucket_versioning(Bucket=bucket_name, VersioningConfiguration=vconfig)
        logging.debug(vresp)
        return web.HTTPCreated()
    except ClientError as e:
        logging.error(e)
        lock_msg = e.response['Error']['Message']

        return web.HTTPBadRequest(body=f"Failed to update versioning status. {lock_msg}")


async def put_bucket_tags(request: Request, volume_id: str, bucket_name: str,
                          new_tags: List[Tag]) -> web.Response:
    """
    Creates or adds to a tag list for bucket

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param bucket_name: Bucket to create
    :param new_tags: new tags to be added tag list on specified bucket
    """
    if not new_tags:
        return web.HTTPNotModified()
    aws_new_tags = await _to_aws_tags(new_tags)

    if not bucket_name:
        return web.HTTPBadRequest(body="Bucket name is required")
    try:
        s3_resource = await _get_resource(request, 's3', volume_id)
        bucket_tagging = s3_resource.BucketTagging(bucket_name)
        tags = []

        try:
            tags = bucket_tagging.tag_set
        except ClientError as ce:
            if ce.response['Error']['Code'] != 'NoSuchTagSet':
                logging.error(ce)
                raise ce
        tags = tags + aws_new_tags
        # boto3 tagging.put only accepts dictionaries of Key Value pairs(Tags)
        bucket_tagging.put(Tagging={'TagSet': tags})
        return web.HTTPCreated()
    except (ClientError, Exception) as e:
        logging.error(e)
        return web.HTTPBadRequest(body=e.response)


async def post_folder(request: Request, volume_id: str, path_name: str):
    """
    Adds a folder to bucket given in parameter, checks that it is the first, if already exists errors are thrown

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param path_name: (str) path to delete folder, split into bucket and folder for function, input as bucket and key values together
    """
    bucket_name = path_name.partition("/")[0]
    folder_name = path_name.partition("/")[2]
    s3_client = await _get_client(request, 's3', volume_id)
    try:
        response = s3_client.head_object(Bucket=bucket_name,
                                         Key=(folder_name + '/'))  # check if folder exists, if not throws an exception
        logging.info(response)
        return web.HTTPBadRequest(body="folder already exists")
    except ClientError as e:
        error_code = e.response['Error']['Code']
        if error_code == '404':  # folder doesn't exist
            add_folder_result = s3_client.put_object(Bucket=bucket_name, Key=(folder_name + '/'))
            logging.info(add_folder_result)
            return web.HTTPCreated()
        else:
            logging.error(e)
            return web.HTTPBadRequest()


async def post_object(request: Request):
    """Upload a file to an S3 bucket, checks that it is the first, if already exists errors are thrown
    query params:
    file_path (str) path to file on local disk (required),
    path_name (str) path for object to be post on bucket (required),
    object_name (str) name of object to be post on bucket, if not provided name will be extracted from file_path (optional)

    :param request: the aiohttp Request (required).
    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-uploading-files.html
    """
    # If S3 object_name was not specified, use file_name
    volume_id = request.match_info.get('volume_id', None)
    bucket_name = request.match_info.get('bucket_name', None)

    if not volume_id:
        return web.HTTPBadRequest(body="volume_id is required")
    if not bucket_name:
        return web.HTTPBadRequest(body="bucket_name is required")

    query_keys = {}
    if request.rel_url and request.rel_url.query:
        for key in request.rel_url.query:
            query_keys[key] = request.rel_url.query[key]
    if 'file_path' not in query_keys or 'path_name' not in query_keys:
        return web.HTTPBadRequest()
    path_name = query_keys['path_name']
    file_path = query_keys['file_path']

    if 'object_name' not in query_keys:
        object_name = os.path.basename(
            (os.path.normpath(file_path)))  # only gets the last part of the path so identifiable info not included
    else:
        object_name = query_keys['object_name']

    s3_client = await _get_client(request, 's3', volume_id)

    try:
        upload_response = s3_client.head_object(Bucket=bucket_name,
                                                Key=path_name + object_name)  # check if folder exists, if not throws an exception
        logging.info(upload_response)
        return web.HTTPBadRequest(body="object already exists")
    except ClientError as e:
        error_code = e.response['Error']['Code']
        if error_code == '404':  # folder doesn't exist
            upload_response = s3_client.upload_file(file_path, bucket_name, path_name + object_name)
            logging.info(upload_response)
            return web.HTTPCreated()
        else:
            logging.info(e)
            return web.HTTPBadRequest()


def post_object_archive():
    """
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/glacier.html
    """


async def put_bucket(request: Request, volume_id: str, bucket_name: str, region: Optional[str] = None):
    """
    Create an S3 bucket in a specified region, if it doesn't exist an error will be thrown
    If a region is not specified, the bucket is created in the S3 default region (us-east-1).

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param bucket_name: Bucket to create
    :param region: String region to create bucket in, e.g., 'us-west-2'
    """
    s3_client = await _get_client(request, 's3', volume_id)
    try:
        s3_client.head_bucket(Bucket=bucket_name)  # check if bucket exists, if not throws an exception
        if region is None:
            create_bucket_result = s3_client.create_bucket(Bucket=bucket_name)
            logging.info(create_bucket_result)
            return web.HTTPCreated()
        else:
            create_bucket_result = s3_client.create_bucket(Bucket=bucket_name,
                                                           CreateBucketConfiguration={'LocationConstraint': region})
            logging.info(create_bucket_result)
            return web.HTTPCreated()
    except ClientError as e:
        error_code = e.response['Error']['Code']
        if error_code == '404':
            logging.error(e)
            return web.HTTPBadRequest(body="bucket doesn't exist")
        elif error_code == '403':
            logging.error(e)
            return web.HTTPBadRequest(body="bucket exists, no permission to access")
        else:
            logging.error(e)
            return web.HTTPBadRequest()


async def put_folder(request: Request, volume_id: str, path_name: str):
    """
    Adds a folder to bucket given in parameter, if it doesn't exist, throws error

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param path_name: (str) path to delete folder, split into bucket and folder for function, input as bucket and key values together
    """
    bucket_name = path_name.partition("/")[0]
    folder_name = path_name.partition("/")[2]
    s3_client = await _get_client(request, 's3', volume_id)
    try:
        s3_client.head_object(Bucket=bucket_name,
                              Key=(folder_name + '/'))  # check if folder exists, if not throws an exception
        add_folder_result = s3_client.put_object(Bucket=bucket_name, Key=(folder_name + '/'))
        logging.info(add_folder_result)
        return web.HTTPCreated()
    except ClientError as e:
        error_code = e.response['Error']['Code']
        if error_code == '404':  # folder doesn't exist
            logging.error(e)
            return web.HTTPBadRequest()
        else:
            logging.error(e)
            return web.HTTPBadRequest()


async def put_object(request: Request, volume_id: str, path_name: str, file_path: str,
                     object_name: Optional[str] = None):
    """Upload a file to an S3 bucket, if doesn't exist, throws error

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param file_path: Path to the file to upload
    :param path_name: path to the location of object
    :param object_name: S3 object name. If not specified then file_name is used

    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-uploading-files.html
    """
    # If S3 object_name was not specified, use file_name
    if object_name is None:
        object_name = os.path.basename(
            (os.path.normpath(file_path)))  # only gets the last part of the path so identifiable info not included

    s3_client = await _get_client(request, 's3', volume_id)
    bucket_name = path_name.partition("/")[0]
    folder_name = path_name.partition("/")[2]
    try:
        s3_client.head_object(Bucket=bucket_name,
                              Key=folder_name + object_name)  # check if folder exists, if not throws an exception
        upload_response = s3_client.upload_file(file_path, bucket_name, folder_name + object_name)
        logging.info(upload_response)
        return web.HTTPCreated()
    except ClientError as e:
        error_code = e.response['Error']['Code']
        if error_code == '404':  # folder doesn't exist
            logging.error(e)
            return web.HTTPBadRequest()
        else:
            logging.error(e)
            return web.HTTPBadRequest()


def put_object_archive():
    """
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/glacier.html
    """


async def transfer_object_within_account(request: Request, volume_id: str, object_path, new_path):
    """
    same as copy_object, but also deletes the object

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param object_path (str) gives the old location of the object, input as the bucket and key together
    :param new_path: (str) gives the new location to put the object
    """
    await copy_object(request, volume_id, object_path, new_path)
    await delete_object(request, volume_id, object_path)


def transfer_object_between_account():
    """
    https://markgituma.medium.com/copy-s3-bucket-objects-across-separate-aws-accounts-programmatically-323862d857ed
    """
    # TODO: use update_bucket_policy to set up "source" bucket policy correctly
    """
    {
    "Version": "2012-10-17",
    "Id": "Policy1546558291129",
    "Statement": [
        {
            "Sid": "Stmt1546558287955",
            "Effect": "Allow",
            "Principal": {
                "AWS": "arn:aws:iam::<AWS_IAM_USER>"
            },
            "Action": [
              "s3:ListBucket",
              "s3:GetObject"
            ],
            "Resource": "arn:aws:s3:::<SOURCE_BUCKET>/",
            "Resource": "arn:aws:s3:::<SOURCE_BUCKET>/*"
        }
    ]
    }
    """
    # TODO: use update_bucket_policy to set up aws "destination" bucket policy
    """
    {
    "Version": "2012-10-17",
    "Id": "Policy22222222222",
    "Statement": [
        {
            "Sid": "Stmt22222222222",
            "Effect": "Allow",
            "Principal": {
                "AWS": [
                  "arn:aws:iam::<AWS_IAM_DESTINATION_USER>",
                  "arn:aws:iam::<AWS_IAM_LAMBDA_ROLE>:role/
                ]
            },
            "Action": [
                "s3:ListBucket",
                "s3:PutObject",
                "s3:PutObjectAcl"
            ],
            "Resource": "arn:aws:s3:::<DESTINATION_BUCKET>/",
            "Resource": "arn:aws:s3:::<DESTINATION_BUCKET>/*"
        }
    ]
    }
    """
    # TODO: code
    source_client = boto3.client('s3', "SOURCE_AWS_ACCESS_KEY_ID", "SOURCE_AWS_SECRET_ACCESS_KEY")
    source_response = source_client.get_object(Bucket="SOURCE_BUCKET", Key="OBJECT_KEY")
    destination_client = boto3.client('s3', "DESTINATION_AWS_ACCESS_KEY_ID", "DESTINATION_AWS_SECRET_ACCESS_KEY")
    destination_client.upload_fileobj(source_response['Body'], "DESTINATION_BUCKET",
                                      "FOLDER_LOCATION_IN_DESTINATION_BUCKET")


async def rename_object(request: Request, volume_id: str, object_path: str, new_name: str):
    """
    BOTO3, the copy and rename is the same
    https://medium.com/plusteam/move-and-rename-objects-within-an-s3-bucket-using-boto-3-58b164790b78
    https://stackoverflow.com/questions/47468148/how-to-copy-s3-object-from-one-bucket-to-another-using-python-boto3

    :param request: the aiohttp Request (required).
    :param volume_id: the id string of the volume representing the user's AWS account.
    :param object_path: (str) path to object, includes both bucket and key values
    :param new_name: (str) value to rename the object as, will only replace the name not the path. Use transfer object for that
    """
    # TODO: check if ACL stays the same and check existence
    try:
        s3_resource = await _get_resource(request, 's3', volume_id)
        copy_source = {'Bucket': object_path.partition("/")[0], 'Key': object_path.partition("/")[2]}
        bucket_name = object_path.partition("/")[0]
        old_name = object_path.rpartition("/")[2]
        s3_resource.meta.client.copy(copy_source, bucket_name,
                                     object_path.partition("/")[2].replace(old_name, new_name))
    except ClientError as e:
        logging.error(e)


def update_bucket_policy():
    """

    """


async def _get_client(request: Request, service_name: str, volume_id: str) -> Any:
    """
    Gets an AWS service client.  If the volume has no credentials, it uses the boto3 library to try and find them.

    :param request: the HTTP request (required).
    :param service_name: AWS service name (required).
    :param volume_id: the id string of a volume (required).
    :return: a Mongo client for the file system specified by the volume's file_system_name attribute. If no volume_id
    was provided, the return value will be the "default" Mongo client for the microservice found in the HEA_DB
    application-level property.
    :raise ValueError: if there is no volume with the provided volume id, the volume's file system does not exist,
    the volume's credentials were not found, or a necessary service is not registered.
    """

    if volume_id is not None:
        file_system, credentials = await get_file_system_and_credentials_from_volume(request, volume_id, AWSFileSystem)
        if credentials is None:
            return boto3.client(service_name)
        else:
            return boto3.client(service_name, aws_access_key_id=credentials.account,
                                aws_secret_access_key=credentials.password)
    else:
        raise ValueError('volume_id is required')


async def _get_resource(request: Request, service_name: str, volume_id: str) -> Any:
    """
    Gets an AWS resource. If the volume has no credentials, it uses the boto3 library to try and find them.

    :param request: the HTTP request (required).
    :param service_name: AWS service name (required).
    :param volume_id: the id string of a volume (required).
    :return: a Mongo client for the file system specified by the volume's file_system_name attribute. If no volume_id
    was provided, the return value will be the "default" Mongo client for the microservice found in the HEA_DB
    application-level property.
    :raise ValueError: if there is no volume with the provided volume id, the volume's file system does not exist,
    the volume's credentials were not found, or a necessary service is not registered.
    """

    if volume_id is not None:
        file_system, credentials = await get_file_system_and_credentials_from_volume(request, volume_id, AWSFileSystem)
        if credentials is None:
            return boto3.resource(service_name)
        else:
            return boto3.resource(service_name, aws_access_key_id=credentials.account,
                                  aws_secret_access_key=credentials.password)
    else:
        raise ValueError('volume_id is required')


async def _get_account(request: Request, volume_id: str) -> Optional[DesktopObjectDict]:
    """
    Gets the current user's AWS account dict associated with the provided volume_id.

    :param request: the HTTP request object (required).
    :param volume_id: the volume id (required).
    :return: the AWS account dict, or None if not found.
    """
    aws_object_dict = {}
    sts_client = await _get_client(request, 'sts', volume_id)
    iam_client = await _get_client(request, 'iam', volume_id)

    aws_object_dict['account_id'] = sts_client.get_caller_identity().get('Account')
    # aws_object_dict['alias'] = next(iam_client.list_account_aliases()['AccountAliases'], None)  # Only exists for IAM accounts.
    user = iam_client.get_user()['User']
    # aws_object_dict['account_name'] = user.get('UserName')  # Only exists for IAM accounts.
    aws_object_dict['id'] = user['UserId']
    aws_object_dict['owner'] = request.headers.get(SUB, NONE_USER)
    aws_object_dict['created'] = user['CreateDate']
    return aws_object_dict


async def _from_aws_tags(aws_tags: List[Dict[str, str]]) -> List[Tag]:
    """
    :param aws_tags: Tags obtained from boto3 Tags api
    :return: List of HEA Tags
    """
    hea_tags = []
    for t in aws_tags:
        tag = Tag()
        tag.key = t['Key']
        tag.value = t['Value']
        hea_tags.append(tag)
    return hea_tags


async def _get_bucket(volume_id: str, s3_resource: boto3.resource, s3_client: boto3.client,
                      bucket_name: Optional[str] = None, bucket_id: Optional[str] = None,
                      creation_date: Optional[str] = None) -> AWSBucket:
    """
    :param volume_id: the volume id
    :param s3_resource: the boto3 resource
    :param s3_client:  the boto3 client
    :param bucket_name: str the bucket name (optional)
    :param bucket_id: str the bucket id (optional)
    :param creation_date: str the bucket creation date (optional)
    :return: Returns either the AWSBucket or None for Not Found or Forbidden, else raises ClientError
    """
    try:

        if not volume_id or (not bucket_id and not bucket_name):
            raise ValueError("volume_id is required and either bucket_name or bucket_id")
        # id_type = 'id' if bucket_id else 'name'
        # user = request.headers.get(SUB)
        # bucket_dict = await request.app[HEA_DB].get(request, MONGODB_BUCKET_COLLECTION, var_parts=id_type, sub=user)
        # if not bucket_dict:
        #     return web.HTTPBadRequest()

        b = AWSBucket()
        b.name = bucket_id if bucket_id else bucket_name
        b.id = bucket_id if bucket_id else bucket_name
        b.display_name = bucket_id if bucket_id else bucket_name

        b.created = creation_date if creation_date else s3_resource.Bucket(b.name).creation_date
        bucket_versioning = s3_resource.BucketVersioning(b.name)
        b.versioned = bucket_versioning.status == 'Enabled'
        b.s3_uri = "s3://" + b.name + "/"
        b.region = s3_client.get_bucket_location(Bucket=b.name)['LocationConstraint']
        # todo how to find partition dynamically. The format is arn:PARTITION:s3:::NAME-OF-YOUR-BUCKET
        # b.arn = "arn:"+"aws:"+":s3:::"
        try:
            t_list = s3_client.get_bucket_tagging(Bucket=b.name)['TagSet']
            b.tags = await _from_aws_tags(aws_tags=t_list)
        except ClientError as ce:
            if ce.response['Error']['Code'] != 'NoSuchTagSet':
                logging.info(ce)

        try:
            encrypt = s3_client.get_bucket_encryption(Bucket=b.name)
            rules: list = encrypt['ServerSideEncryptionConfiguration']['Rules']
            b.encrypted = len(rules) > 0
        except ClientError as e:
            if e.response['Error']['Code'] == 'ServerSideEncryptionConfigurationNotFoundError':
                b.encrypted = False
            else:
                raise e
        try:
            b.permission_policy = s3_client.get_bucket_policy(Bucket=b.name)['Policy']
        except ClientError as e:
            if e.response['Error']['Code'] != 'NoSuchBucketPolicy':
                logging.error(e)
                raise e
        try:
            lock_config = s3_client.get_object_lock_configuration(Bucket=b.name)
            b.locked = lock_config['ObjectLockConfiguration']['ObjectLockEnabled'] == 'Enabled'
        except ClientError as e:
            if e.response['Error']['Code'] != 'ObjectLockConfigurationNotFoundError':
                logging.error(e)
            b.locked = False
        # todo need to lazy load this these metrics
        total_size = None
        obj_count = None
        mod_date = None
        # FIXME need to calculate this metric data in a separate call. Too slow
        # s3bucket = s3_resource.Bucket(b.name)
        # for obj in s3bucket.objects.all():
        #     total_size += obj.size
        #     obj_count += 1
        #     mod_date = obj.last_modified if mod_date is None or obj.last_modified > mod_date else mod_date
        b.size = total_size
        b.object_count = obj_count
        b.modified = mod_date

    except ClientError as e:
        error_code = e.response['Error']['Code']
        if error_code == '403' or error_code == '404':
            return None
        logging.error(e)
        raise e
    return b


async def _to_aws_tags(hea_tags: List[Tag]) -> List[Dict[str, str]]:
    """
    :param hea_tags: HEA tags to converted to aws tags compatible with boto3 api
    :return: aws tags
    """
    aws_tag_dicts = []
    for hea_tag in hea_tags:
        aws_tag_dict = {}
        aws_tag_dict['Key'] = hea_tag.key
        aws_tag_dict['Value'] = hea_tag.value
        aws_tag_dicts.append(aws_tag_dict)
    return aws_tag_dicts

# if __name__ == "__main__":
# print(get_account())
# print(post_bucket('richardmtest'))
# print(put_bucket("richardmtest"))
# print(get_all_buckets())
# print(get_all("richardmtest"))
# print(post_folder('richardmtest/temp'))
# print(put_folder("richardmtest/temp"))
# print(post_object(r'richardmtest/temp/', r'C:\Users\u0933981\IdeaProjects\heaserver\README.md'))
# print(put_object(r'richardmtest/temp/', r'C:\Users\u0933981\IdeaProjects\heaserver\README.md'))
# download_object(r'richardmtest/temp/README.md', r'C:\Users\u0933981\Desktop\README.md')
# rename_object(r'richardmtest/README.md', 'readme2.md')
# print(copy_object(r'richardmtest/temp/README.md', r'richardmtest/temp/README.md'))
# print(transfer_object_within_account(r'richardmtest/temp/readme2.md', r'timmtest/temp/README.md'))
# print(generate_presigned_url(r'richardmtest/temp/'))
# print(get_object_content(r'richardmtest/temp/README.md'))  # ["Body"].read())
# print(delete_object('richardmtest/temp/README.md'))
# print(delete_folder('richardmtest/temp'))
# print(delete_bucket_objects("richardmtest"))
# print(delete_bucket('richardmtest'))
# print("done")



from aiohttp.web import Application
from .mongotestcase import MongoTestCase
from ..db.mongo import Mongo, replace_id_with_object_id
from .. import runner
from .. import wstl
from ..docker import DockerImages
import logging
from typing import Union, Dict, List, Any, Optional, Type, Iterable
from testcontainers.mongodb import MongoDbContainer
import gridfs
from bson import ObjectId
from contextlib import ExitStack
from io import BytesIO
from yarl import URL
import os
from .expectedvalues import expected_values, ActionSpec, LinkSpec
from heaserver.service.docker import get_exposed_port, start_microservice_container, DockerContainer
from array import array
from heaobject.user import NONE_USER


def get_test_case_cls(href: Union[URL, str],
                      wstl_package: str,
                      coll: str,
                      fixtures: Dict[str, List[dict]],
                      content: Optional[Dict[str, Dict[str, bytes]]] = None,
                      content_type: Optional[str] = None,
                      put_content_status: Optional[int] = None,
                      body_post: Optional[Dict[str, Dict[str, Any]]] = None,
                      body_put: Optional[Dict[str, Dict[str, Any]]] = None,
                      expected_one: Optional[Dict[str, Any]] = None,
                      expected_one_wstl: Optional[Dict[str, Any]] = None,
                      expected_one_duplicate_form: Optional[Dict[str, Any]] = None,
                      expected_all: Optional[Dict[str, Any]] = None,
                      expected_all_wstl: Optional[Dict[str, Any]] = None,
                      expected_opener: Optional[Union[str, URL]] = None,
                      expected_opener_body: Optional[Dict[str, Any]] = None,
                      registry_docker_image: Optional[str] = None,
                      port: Optional[int] = None,
                      sub: Optional[str] = NONE_USER) -> Type[MongoTestCase]:
    """
    This function configures mocks. It returns a base test case class for testing a mongodb-based service with the
    provided fixtures. The test case class is a subclass of aiohttp.test_utils.AioHTTPTestCase, and the provided
    fixtures can be found in the following fields: _resource_path, _body_post, _body_put, _expected_all, and
    _expected_one.

    :param href: the resource being tested (required).
    :param wstl_package: the name of the package containing the wstl data package (required).
    :param coll: the MongoDB collection (required).
    :param fixtures: data to insert into the mongo database before running each test. Should be a dict of
    collection -> list of objects. Required.
    :param content: :param content: data to insert into GridFS (optional), as a collection string -> HEA Object id ->
    content as a bytes, bytearray, or array.array object.
    :param content_type: the MIME type of the content (optional).
    :param put_content_status: the expected status code for updating the content of the HEA object (optional). Normally
    should be 204 if the content is updatable and 405 if not. Default is None, which will cause associated tests to be
    skipped.
    :param body_post: JSON dict for data to be posted.
    :param body_put: JSON dict for data to be put. If None, the value of body_post will be used for PUTs.
    :param expected_one: The expected JSON dict for GET calls. If None, the value of expected_all will be used.
    :param expected_one_wstl: The expected JSON dict for GET calls that return the
    application/vnd.wstl+json mime type.
    :param expected_one_duplicate_form: The expected JSON dict for GET calls that return the
    object's duplicate form.
    :param expected_all: The expected JSON dict for GET-all calls.
    :param expected_all_wstl: The expected JSON dict for GET-all calls that return the
    application/vnd.wstl+json mime type.
    :param expected_opener: The expected URL of the resource that does the opening.
    :param expected_opener_body: The expected JSON dict for GET calls for an HEA desktop object's opener choices.
    :param registry_docker_image: an heaserver-registry docker image in REPOSITORY:TAG format, that will be launched
    after the MongoDB container is live.
    :param port: the port number to run aiohttp. If None, a random available port will be chosen.
    :return the base test case class.
    """

    class RealMongoTestCase(MongoTestCase):
        """
        Test case class for testing a mongodb-based service.
        """

        def __init__(self, methodName: str = 'runTest') -> None:
            """
            Initializes a test case.

            :param methodName: the name of the method to test.
            """
            super().__init__(methodName=methodName,
                             href=href,
                             coll=coll,
                             body_post=body_post,
                             body_put=body_put,
                             expected_all=expected_all,
                             expected_one=expected_one or expected_all,
                             expected_one_wstl=expected_one_wstl or expected_all_wstl,
                             expected_all_wstl=expected_all_wstl,
                             expected_one_duplicate_form=expected_one_duplicate_form,
                             expected_opener=expected_opener,
                             expected_opener_body=expected_opener_body,
                             wstl_package=wstl_package,
                             content=content,
                             content_type=content_type,
                             put_content_status=put_content_status,
                             port=port,
                             sub=sub)

        def run(self, result=None):
            """
            Runs a test using a freshly created MongoDB Docker container. The container is destroyed upon concluding
            the test.

            :param result: a TestResult object into which the test's result is collected.
            :return: the TestResult object.
            """
            with ExitStack() as stack, self._caplog.at_level(logging.DEBUG):
                config_file = setup_test_environment(stack, fixtures, content, registry_docker_image=registry_docker_image)
                self.__config = runner.init(config_string=config_file)
                return super().run(result)

        async def get_application(self) -> Application:
            return runner.get_application(db=Mongo,
                                          wstl_builder_factory=wstl.builder_factory(wstl_package, href=href),
                                          config=self.__config,
                                          testing=True)

    return RealMongoTestCase


def get_test_case_cls_default(coll: str,
                              wstl_package: str,
                              fixtures: Dict[str, List[Dict[str, Any]]],
                              duplicate_action_name: str,
                              content: Optional[Dict[str, Dict[str, bytes]]] = None,
                              content_type: Optional[str] = None,
                              put_content_status: Optional[int] = None,
                              include_root=False,
                              href: Optional[Union[str, URL]] = None,
                              get_actions: Optional[List[ActionSpec]] = None,
                              get_all_actions: Optional[List[ActionSpec]] = None,
                              expected_opener: Optional[LinkSpec] = None,
                              registry_docker_image: Optional[str] = None,
                              port: Optional[int] = None,
                              sub: Optional[str] = NONE_USER) -> Type[MongoTestCase]:
    if href is None:
        href_ = str(URL(f'/{coll}/'))
    else:
        href_ = str(href)
        if not href_.endswith('/'):
            href_ = href_ + '/'
    return get_test_case_cls(href=href_, wstl_package=wstl_package, coll=coll, fixtures=fixtures,
                             content=content, content_type=content_type, put_content_status=put_content_status,
                             **expected_values(fixtures, coll, wstl.builder(package=wstl_package),
                                               duplicate_action_name, href_,
                                               include_root=include_root,
                                               get_actions=get_actions, get_all_actions=get_all_actions,
                                               opener_link=expected_opener),
                             registry_docker_image=registry_docker_image,
                             port=port, sub=sub
                             )


def setup_test_environment(stack: ExitStack,
                           fixtures: Dict[str, List[dict]],
                           content: Optional[Dict[str, Dict[str, bytes]]],
                           registry_docker_image: Optional[str] = None,
                           other_docker_images: Optional[Iterable[DockerContainer]] = None) -> str:
    _logger = logging.getLogger(__name__)
    os.environ['MONGO_DB'] = 'hea'
    mongo_ = stack.enter_context(MongoDbContainer(DockerImages.MONGODB.value))

    db = mongo_.get_connection_client().hea
    for k in fixtures or {}:
        db[k].insert_many(replace_id_with_object_id(f) for f in fixtures[k])
    content_ = content if content is not None else {}
    for k in content_:
        fs = gridfs.GridFSBucket(db, bucket_name=k)
        for id_, d in content_[k].items():
            if isinstance(d, (bytes, bytearray, array)):
                with BytesIO(d) as b:
                    fs.upload_from_stream_with_id(ObjectId(id_), id_, b)

    mongodb_connection_string = f'mongodb://test:test@{mongo_.get_container_host_ip()}:{get_exposed_port(mongo_, 27017)}/hea?authSource=admin'
    if registry_docker_image is None:
        config_file = f"""
                                        [MongoDB]
                                        ConnectionString = {mongodb_connection_string}
                                        """
    else:
        if other_docker_images is None:
            other_docker_images_: Iterable[DockerContainer] = []
        else:
            other_docker_images_ = other_docker_images
        registry_url = start_microservice_container(DockerContainer(image=registry_docker_image, port=8080, check_path='/components'), mongo_, stack)
        for image in other_docker_images_:
            start_microservice_container(image, mongo_, stack)
        config_file = f"""
                                        [DEFAULT]
                                        Registry={registry_url}

                                        [MongoDB]
                                        ConnectionString = {mongodb_connection_string}
                                            """
    _logger.debug('HEA config file: %s', config_file)
    return config_file



from aiohttp import hdrs, web
from heaobject import root
from heaobject.root import DesktopObject, DesktopObjectTypeVar, desktop_object_type_for_name
from heaobject.registry import Component, Property
from heaobject.volume import DEFAULT_FILE_SYSTEM, DefaultFileSystem, FileSystem
from . import appproperty, response
from .aiohttp import StreamReaderWrapper, SupportsAsyncRead
from .representor import nvpjson
from yarl import URL
from typing import Optional, Union, Dict, Type, Mapping
import logging


async def get_streaming(request: web.Request, url: Union[str, URL], headers: Optional[Dict[str, str]] = None) -> web.StreamResponse:
    _logger = logging.getLogger(__name__)
    session = request.app[appproperty.HEA_CLIENT_SESSION]
    _logger.debug('Getting streaming content at %s with headers %s', url, headers)
    async with session.get(url, headers=headers, raise_for_status=False) as response_:
        if response_.status == 200:
            content_type = response_.headers.get(hdrs.CONTENT_TYPE, None)
            return await response.get_streaming(request, StreamReaderWrapper(response_.content), content_type=content_type)
        elif response_.status == 404:
            return response.status_not_found()
        else:
            return web.Response(status=response_.status)


async def put_streaming(request: web.Request, url: Union[str, URL], data: SupportsAsyncRead, headers: Optional[Dict[str, str]] = None) -> None:
    session = request.app[appproperty.HEA_CLIENT_SESSION]
    async with session.put(url, data=data, headers=headers) as response_:
        pass


async def get_dict(app, url, headers=None) -> Optional[dict]:
    """
    Co-routine that gets a dict from a HEA service that returns JSON.

    :param app: the aiohttp application context (required).
    :param url: The URL (str or URL) of the resource (required).
    :param headers: optional dict of headers.
    :return: the dict populated with the resource's content, None if no such resource exists, or another HTTP
    status code if an error occurred.
    """
    _logger = logging.getLogger(__name__)
    session = app[appproperty.HEA_CLIENT_SESSION]
    _logger.debug('Getting dict at %s with headers %s', url, headers)
    async with session.get(url, headers=headers, raise_for_status=False) as response_:
        if response_.status == 404:
            return None
        else:
            response_.raise_for_status()
            result = await response_.json()
            _logger.debug('Client returning %s', result)
            result_len = len(result);
            if result_len != 1:
                raise ValueError(f'Result from {url} has {result_len} values')
            return result[0] if isinstance(result, list) else result


async def get(app: web.Application, url: Union[URL, str], obj: Union[str, DesktopObjectTypeVar, Type[DesktopObjectTypeVar]], headers: Optional[Mapping[str, str]] = None) -> Optional[DesktopObjectTypeVar]:    # type: ignore[return]
    """
    Co-routine that gets a HEA desktop object from a HEA service.

    :param app: the aiohttp application context (required).
    :param url: The URL (str or URL) of the resource (required).
    :param obj: the HEA desktop object type to populate with the resource's content, or a desktop object instance. If a type or type
    name, this function will attempt to create an instance using the type's no-arg constructor.
    :param headers: optional dict of headers. Attempts to set the Accept header will be ignored. The service will
    always receive Accepts: application/json.
    :return: the HEAObject populated with the resource's content, None if no such resource exists, or another HTTP
    status code if an error occurred.
    """
    _logger = logging.getLogger(__name__)
    if isinstance(obj, str):
        obj__ = root.type_for_name(obj)
        if isinstance(obj__, root.DesktopObject):
            obj = obj__
        else:
            raise ValueError(f'Type {obj} is not a DesktopObject')
    if isinstance(obj, type) and issubclass(obj, root.DesktopObject):
        obj_ = obj()
    elif isinstance(obj, root.DesktopObject):
        obj_ = obj
    else:
        raise TypeError('obj must be an HEAObject instance, an HEAObject type, or an HEAObject type name')
    headers_ = dict(headers) if headers else {}
    headers_[hdrs.ACCEPT] = nvpjson.MIME_TYPE

    session = app[appproperty.HEA_CLIENT_SESSION]
    _logger.debug('Getting content at %s with headers %s', url, headers_)
    async with session.get(url, headers=headers_, raise_for_status=False) as response_:
        if response_.status == 404:
            return None
        else:
            response_.raise_for_status()
            result = await response_.json()
            _logger.debug('Client returning %s', result)
            result_len = len(result);
            if result_len != 1:
                raise ValueError(f'Result from {url} has {result_len} values')
            obj_.from_dict(result[0])
            return obj_


async def get_all(app: web.Application, url: Union[URL, str], obj: Type[DesktopObject], headers: Optional[Dict[str, str]] = None):    # type: ignore[return]
    """
    Generator that returns the requested HEAObjects.

    :param app: the aiohttp application context (required).
    :param url: The URL (str or URL) of the resource (required).
    :param obj: the HEAObject type to populate with the resource's content.
    :param headers: optional dict of headers. Attempts to set the Accept header will be ignored. The service will
    always receive Accepts: application/json.
    """
    _logger = logging.getLogger(__name__)
    if isinstance(obj, str):
        obj = root.type_for_name(obj)
    if isinstance(obj, type) and issubclass(obj, root.DesktopObject):
        obj_ = obj
    else:
        raise TypeError('obj must be an HEAObject instance, an HEAObject type, or an HEAObject type name')
    headers_ = dict(headers) if headers else {}
    headers_[hdrs.ACCEPT] = nvpjson.MIME_TYPE

    session = app[appproperty.HEA_CLIENT_SESSION]
    _logger.debug('Getting content at %s with headers %s', url, headers_)
    async with session.get(url, headers=headers_, raise_for_status=False) as response_:
        response_.raise_for_status()
        result = await response_.json()
        _logger.debug('Client returning %s', result)
        for r in result:
            obj__ = obj_()
            obj__.from_dict(r)
            yield obj__


async def post(app: web.Application, url: Union[URL, str], data: root.HEAObject, headers: Dict[str, str] = None) -> str:    # type: ignore[return]
    """
    Coroutine that posts a HEAObject to a HEA service.

    :param app: the aiohttp application context (required).
    :param url: the The URL (str or URL) of the resource (required).
    :param data: the HEAObject (required).
    :param headers: optional dict of headers.
    :return: the URL string in the response's Location header.
    """
    session = app[appproperty.HEA_CLIENT_SESSION]
    async with session.post(url, json=data, headers=headers) as response_:
        return response_.headers['Location']


async def put(app: web.Application, url: Union[URL, str], data: root.HEAObject, headers: Dict[str, str] = None) -> None:
    """
    Coroutine that updates a HEAObject.

    :param app: the aiohttp application context (required).
    :param url: the The URL (str or URL) of the resource (required).
    :param data: the HEAObject (required).
    :param headers: optional dict of headers.
    """
    session = app[appproperty.HEA_CLIENT_SESSION]
    async with session.put(url, json=data, headers=headers) as response_:
        pass


async def delete(app: web.Application, url: Union[URL, str], headers: Dict[str, str] = None) -> None:
    """
    Coroutine that deletes a HEAObject.

    :param app: the aiohttp application context (required).
    :param url: the URL (str or URL) of the resource (required).
    :param headers: optional dict of headers.
    """
    _logger = logging.getLogger(__name__)
    session = app[appproperty.HEA_CLIENT_SESSION]
    _logger.debug('Deleting %s', str(url))
    async with session.delete(url, headers=headers) as response_:
        pass


async def get_component_by_name(app: web.Application, name: str) -> Optional[Component]:
    """
    Gets the Component with the given name from the HEA registry service.

    :param app: the aiohttp app.
    :param name: the component's name.
    :return: a Component instance or None (if not found).
    """
    return await get(app, URL(app[appproperty.HEA_REGISTRY]) / 'components' / 'byname' / name, Component)


async def get_resource_url(app: web.Application, type_or_type_name: Union[str, Type[root.HEAObject]],
                           file_system_type_or_type_name: Union[str, Type[FileSystem]] = DefaultFileSystem,
                           file_system_name: str = DEFAULT_FILE_SYSTEM) -> Optional[str]:
    """
    Gets the resource URL corresponding to the HEA object type from the HEA registry service.

    :param app: the aiohttp app.
    :param type_or_type_name: the HEAObject type or type name of the resource.
    :param file_system_type_or_type_name: optional file system type or type name. The default is heaobject.volume.DefaultFileSystem.
    :param file_system_name: optional file system name. The default is heaobject.volume.FileSystem.DEFAULT_FILE_SYSTEM.
    :return: a URL string or None (if not found).
    """
    if file_system_name is None:
        file_system_name_ = DEFAULT_FILE_SYSTEM
    else:
        file_system_name_ = file_system_name
    if file_system_type_or_type_name is None:
        file_system_type_: Type[DesktopObject] = DefaultFileSystem
    elif isinstance(file_system_type_or_type_name, type):
        if issubclass(file_system_type_or_type_name, FileSystem):
            file_system_type_ = file_system_type_or_type_name
        else:
            raise TypeError('file_system_type_or_type_name not a FileSystem')
    else:
        file_system_type_ = desktop_object_type_for_name(file_system_type_or_type_name)
        if not issubclass(file_system_type_, FileSystem):
            raise TypeError('file_system_type_or_type_name not a FileSystem')
    type_ = _handle_type_or_type_name_arg(type_or_type_name)
    url = URL(app[appproperty.HEA_REGISTRY]) / 'components' / 'bytype' / type_ / 'byfilesystemtype'/ file_system_type_.get_type_name() / 'byfilesystemname' / file_system_name_
    component: Optional[Component] = await get(app, url, Component)
    return component.get_resource_url(type_, file_system_name=file_system_name_) if component is not None else None


async def get_property(app: web.Application, name: str) -> Optional[Property]:
    """
    Gets the Property with the given name from the HEA registry service.

    :param app: the aiohttp app.
    :param name: the property's name.
    :return: a Property instance or None (if not found).
    """
    return await get(app, URL(app[appproperty.HEA_REGISTRY]) / 'properties' / 'byname' / name, Property)


def _handle_type_or_type_name_arg(type_or_type_name):
    if isinstance(type_or_type_name, str) and root.is_heaobject_type(type_or_type_name):
        return type_or_type_name
    elif isinstance(type_or_type_name, type) and issubclass(type_or_type_name, root.HEAObject):
        return type_or_type_name.get_type_name()
    else:
        raise TypeError('type_or_type_name must be an HEAObject type')



import json
from urllib.request import urlopen

with urlopen('https://gitlab.com/huntsman-cancer-institute/risr/hea/hea-json-schemas/-/raw/master/wstlaction.json') as url:
    WSTL_ACTION_SCHEMA = json.loads(url.read().decode("utf-8"))

with urlopen('https://gitlab.com/huntsman-cancer-institute/risr/hea/hea-json-schemas/-/raw/master/wstl.json') as url:
    WSTL_SCHEMA = json.loads(url.read().decode("utf-8"))

with urlopen('https://gitlab.com/huntsman-cancer-institute/risr/hea/hea-json-schemas/-/raw/master/cjtemplate.json') as url:
    CJ_TEMPLATE_SCHEMA = json.loads(url.read().decode("utf-8"))

with urlopen('https://gitlab.com/huntsman-cancer-institute/risr/hea/hea-json-schemas/-/raw/master/nvpjson.json') as url:
    NVPJSON_SCHEMA = json.loads(url.read().decode("utf-8"))



"""
This module implements convenience functions for querying databases using the Python Database API Specification v2.0
(DB-API2). Documentation for DB-API2 is available from https://www.python.org/dev/peps/pep-0249/.
"""
import contextlib


@contextlib.contextmanager
def db_connect(db_driver, cn_params):
    """
    Manage a connection to a database.

    :param db_driver: the DB-API2 module (required).
    :param cn_params: a tuple or dict of connection parameters to pass to the driver's connect function (required).
    :return: the connection.
    :raise: any exception thrown by the database driver.
    """
    if isinstance(cn_params, dict):
        cn = db_driver.connect(**cn_params)
    else:
        cn = db_driver.connect(*cn_params)
    try:
        yield cn
        cn.close()
        cn = None
    finally:
        _close_quietly(cn)


@contextlib.contextmanager
def transaction(cn):
    """
    Context manager that automatically commits the current transaction before leaving the context. If an exception is
    raised, it will rollback the current transaction instead. Because DB-API2 only supports implicit transactions,
    transaction start is the first statement executed, or the first statement executed since the most recent commit or
    rollback. Therefore, this context manager does not define a transaction start -- it could continue a transaction
    that was started previously.

    :param cn: the connection (required).
    :return: the connection. Because you already must have the connection, this typically would be called omitting the
    'as' syntax.
    """
    try:
        yield cn
        cn.commit()
    except Exception as e:
        try:
            cn.rollback()
        except:
            pass
        raise e


def execute(cn, stmt, stmt_params=None, handle_result=None):
    """
    Execute the given statement in the database, and process the result in the handle_result method.

    :param cn: a database connection (required).
    :param stmt: the database operation to execute (required).
    :param stmt_params: parameters to bind to variables in the operation.
    :param handle_result: a callable that accepts one cursor argument (optional). The _execute method handles closing
    the cursor for you, so there is no need to close it in this callable.
    :return: the return value from calling handle_result, or None if handle_result is None.
    :raise: any exception thrown by the database driver.
    """
    with contextlib.closing(cn.cursor()) as cur:
        cur.execute(stmt, stmt_params)
        if handle_result:
            return handle_result(cur)
        else:
            return None


def execute_cn(db_driver, cn_params, stmt, stmt_params=None, handle_result=None):
    """
    Create a connection, execute the given statement in the database, and process the result in the handle_result
    method.

    :param db_driver: the DB-API2 module (required).
    :param cn_params: a tuple of connection parameters to pass to the driver's connect function (required).
    :param stmt: the database operation to execute (required).
    :param stmt_params: parameters to bind to variables in the operation.
    :param handle_result: a callable that accepts one cursor argument (optional). The _execute method handles closing
    the cursor for you, so there is no need to close it in this callable.
    :return: the return value from calling handle_result, or None if handle_result is None.
    :raise: any exception thrown by the database driver.
    """
    with db_connect(db_driver, cn_params) as cn:
        return execute(cn, stmt, stmt_params, handle_result)


def execute_many(cn, stmt, stmt_params):
    """
    Execute the given statement in the database many times with different parameters. SELECT statements cannot be
    executed with this function -- use _execute_one instead.

    :param cn: a database connection (required).
    :param stmt: the database operation to execute (required).
    :param stmt_params: a sequence of sequences of parameters to bind to variables in each operation.
    :return:
    """
    with contextlib.closing(cn.cursor()) as cur:
        cur.executemany(stmt, stmt_params)


def execute_many_cn(db_driver, cn_params, stmt, stmt_params=None):
    """
    Create a connection, and execute the given statement in the database many times with different parameters. SELECT
    statements cannot be executed with this function -- use _execute_one instead.

    :param db_driver: the DB-API2 module (required).
    :param cn_params: a tuple or dict of connection parameters to pass to the driver's connect function (required).
    :param stmt: the database operation to execute (required).
    :param stmt_params: a sequence of sequences of parameters to bind to variables in each operation.
    :return: the return value from calling handle_result, or None if handle_result is None.
    :raise: any exception thrown by the database driver.
    """
    with db_connect(db_driver, cn_params) as cn:
        with contextlib.closing(cn.cursor()) as cur:
            cur.executemany(stmt, stmt_params)


def generate(db_driver, cn_params, stmt, next_params, count, commit_freq=10000):
    """
    Generate parameters and use repeatedly in a database DML operation.

    :param db_driver: the DB-API2 module (required).
    :param cn_params: a tuple or dict of connection parameters to pass to the driver's connect function (required).
    :param stmt: the DML operation to execute (required).
    :param next_params: a generator of parameter tuples to include in the database operation.
    :param count: the number of times to call the generator.
    :param commit_freq: the frequency of commits (required; default is every 10,000 rows).
    """
    with db_connect(db_driver, cn_params) as cn:
        id_ = 0
        while id_ < count:
            with transaction(cn):
                locations = tuple(next_params() for i in range(commit_freq) if id_ + i < count)
                id_ += len(locations)
                execute_many(cn, stmt, locations)
        cn.commit()


def _close_quietly(obj):
    """
    If obj is not None, this function calls its close() method ignoring any exceptions that are
    raised. If obj is None, this function does nothing.

    :param obj: an object with a no-argument close method (optional).
    """
    if obj is not None:
        try:
            obj.close()
        except:
            pass



"""
Utility functions and classes for working with docker images and containers.
"""
from contextlib import ExitStack

from testcontainers.general import DockerContainer
from testcontainers.core.waiting_utils import wait_container_is_ready
import requests
from testcontainers.mongodb import MongoDbContainer
from yarl import URL
from typing import Dict, List, Tuple
from copy import deepcopy, copy
from typing import Optional
import logging
from heaobject.registry import Resource
from enum import Enum


class DockerImages(Enum):
    """
    Images to use for SwaggerUI, tests, and other situations in which we want HEA to start Docker containers
    automatically.
    """
    MONGODB = 'percona/percona-server-mongodb:4.4.9'


class DockerVolumeMapping:
    """
    Docker volume mapping. This class is immutable.
    """

    def __init__(self, host: str, container: str, mode: str = 'ro'):
        """
        Creates a volume mapping.

        :param host: the path of the directory on the host to map (required).
        :param container: the path to mount the volume in the container (required).
        :param mode: access level in the container as Unix rwx-style permissions (defaults to 'ro').
        """
        if mode is None:
            self.__mode: str = 'ro'
        else:
            self.__mode = str(mode)
        self.__container = str(container)
        self.__host = str(host)

    @property
    def host(self) -> str:
        return self.__host

    @property
    def container(self) -> str:
        return self.__container

    @property
    def mode(self) -> str:
        return self.__mode


class DockerContainerSpec:
    """
    Docker image and configuration for starting a container. This class is immutable.
    """

    def __init__(self, image: str, port: int, check_path: Optional[str] = None,
                 resources: Optional[List[Resource]] = None,
                 volumes: Optional[List[DockerVolumeMapping]] = None,
                 env_vars: Optional[Dict[str, str]] = None):
        """
        Constructor.

        :param image: the image tag (required).
        :param port: the exposed port (required).
        :param check_path: the URL path to check if the microservice is running.
        :param resources: a list of heaobject.registry.Resource dicts indicating what content types this image is designed for.
        :param volumes: a list of volume mappings.
        :param env_vars: a dict containing environment variable names mapped to string values.
        """
        if image is None:
            raise ValueError('image cannot be None')
        if port is None:
            raise ValueError('port cannot be None')
        if any(not isinstance(volume, DockerVolumeMapping) for volume in volumes or []):
            raise TypeError(f'volumes must contain only {DockerVolumeMapping} objects')
        if any(not isinstance(k, str) and isinstance(v, str) for k, v in (env_vars or {}).items()):
            raise TypeError('env_vars must be a str->str dict')
        if any(not isinstance(r, Resource) for r in resources or []):
            raise TypeError(f'resources must contain only {Resource} objects')
        self.__image = str(image)
        self.__port = int(port)
        self.__check_path = str(check_path)
        self.__resources = [deepcopy(e) for e in resources or []]
        self.__volumes = list(volumes) if volumes else []
        self.__env_vars = dict(env_vars) if env_vars is not None else {}

    @property
    def image(self) -> str:
        """
        The image tag (read-only).
        """
        return self.__image

    @property
    def port(self) -> int:
        """
        The exposed port (read-only).
        """
        return self.__port

    @property
    def check_path(self) -> Optional[str]:
        """
        The URL path to check for whether the microservice is running (read-only).
        """
        return self.__check_path

    @property
    def resources(self) -> Optional[List[Resource]]:
        """
        A list of heaobject.registry.Resource dicts indicating what content types this image is designed for (read-only).
        """
        return deepcopy(self.__resources)

    @property
    def volumes(self) -> List[DockerVolumeMapping]:
        """
        A list of VolumeMapping instances indicating what volumes to map (read-only, never None).
        """
        return copy(self.__volumes)

    @property
    def env_vars(self) -> Dict[str, str]:
        """
        A dict of environment variable names to string values.
        """
        return copy(self.__env_vars)

    def start(self, stack: ExitStack) -> DockerContainer:
        """
        Start a container using this image, port number, and volume mappings, connecting to the provided MongoDB
        container for database access, and using the provided HEA Server Registry instance.

        :param mongo_: the MongoDB container to use (required).
        :param registry_url: the URL of the registry service to use (optional).
        :param stack: the ExitStack to use (required).
        :return: itself, but started.
        """
        container = DockerContainer(self.image)
        for env, val in self.env_vars.items():
            container.with_env(env, val)
        for volume in self.volumes:
            container.with_volume_mapping(volume.host, volume.container, volume.mode)
        container.with_exposed_ports(self.port)
        microservice = stack.enter_context(container)
        return microservice

    def with_env_vars(self, env_vars: Dict[str, str]) -> 'DockerContainerSpec':
        """
        Returns a new DockerContainerSpec with the same values as this one, plus any environment variables in the
        env_vars argument.

        :param env_vars: any environment variables.
        :return:
        """
        new_env_vars = self.env_vars
        new_env_vars.update(env_vars)
        return DockerContainerSpec(self.image, self.port, self.check_path, self.resources, self.volumes, new_env_vars)


@wait_container_is_ready()
def get_exposed_port(container: DockerContainer, port: int) -> int:
    """
    Returns the actual port that the docker container is listening to. It tries getting the port repeatedly until the
    container has sufficiently started to assign the port number.

    :param container: the docker container (required).
    :param port: the port to which the container's application is listening internally.
    :return: the exposed port.
    """
    return container.get_exposed_port(port)


@wait_container_is_ready()
def wait_for_status_code(url, status: int):
    """
    Makes a HTTP GET call to the provided URL repeatedly until the returned status code is equal to the provided code.

    :param url: the URL to call.
    :param status: the status code to check for.
    """
    actual_status = requests.get(url).status_code
    if actual_status != status:
        raise ValueError(f'Expected status {status} and actual status {actual_status}')


@wait_container_is_ready()
def get_bridge_ip(container: DockerContainer) -> str:
    """
    Returns the IP address of the container on the default bridge network.
    :param container: a docker container.
    :return: an IP address.
    """
    return container.get_docker_client().bridge_ip(container.get_wrapped_container().id)


def start_microservice_container(container_spec: DockerContainerSpec, mongo_: MongoDbContainer, stack: ExitStack,
                                 registry_url: Optional[str] = None,
                                 mongo_hea_database: Optional[str] = 'hea',
                                 mongo_hostname: Optional[str] = None,
                                 mongo_hea_username: Optional[str] = 'test',
                                 mongo_hea_password: Optional[str] = 'test') -> Tuple[str, str]:
    """
    Starts a Docker container with the provided HEA microservice image and configuration (container_spec argument),
    Mongo database container, and exit stack for cleaning up resources. If the docker_image object has a check_path,
    the function will wait until the microservice returns a 200 status code from a GET call to the path before
    returning a two-tuple with the container's external and bridge URLs. This function wraps DockerImage's start()
    method.

    The following environment variables are set in the container and will overwrite any pre-existing values that were
    set using the image's env_vars property:
        MONGO_HEA_DATABASE is set to the value of the hea_database argument.
        HEASERVER_REGISTRY_URL is set to the value of the registry_url argument.
        MONGO_HOSTNAME is set to the value of the mongo_hostname argument.
        MONGO_HEA_DATABASE is set to the value of the mongo_hea_database argument.
        MONGO_HEA_USERNAME is set to the value of the mongo_hea_username argument.
        MONGO_HEA_PASSWORD is set to the value of the mongo_hea_password argument.

    Any other environment variables set using the image's env_vars property are retained.

    :param container_spec: the Docker image to start (required).
    :param mongo_: the MongoDBContainer (required)
    :param stack: the ExitStack (required).
    :param registry_url: optional base URL for the heaserver-registry microservice.
    :param mongo_hea_database: optional MongoDB database name to use. The default is 'hea'.
    :param mongo_hostname: optional MongoDB hostname. If None, the bridge IP address of the provided MongoDB container
    is used.
    :param mongo_hea_username: username for accessing the database. The default is 'test'.
    :param mongo_hea_password: password for accessing the database. The default is 'test'.
    :return: a two-duple containing the container's external URL string and the bridge URL string.
    """
    logger = logging.getLogger(__name__)
    logger.debug('Starting docker container %s', container_spec.image)
    image_ = copy(container_spec)
    if mongo_hea_database is None:
        mongo_hea_database = 'hea'
    if mongo_hea_username is None:
        mongo_hea_username = 'test'
    if mongo_hea_password is None:
        mongo_hea_password = 'test'
    if mongo_hostname is None:
        mongo_hostname = get_bridge_ip(mongo_)
    image_ = _with_hea_env_vars(image_, registry_url, mongo_hostname, mongo_hea_database, mongo_hea_username, mongo_hea_password)

    microservice = image_.start(stack)

    external_url = f'http://{microservice.get_container_host_ip()}:{get_exposed_port(microservice, image_.port)}'
    logger.debug('External URL of docker image %s is %s', image_.image, external_url)

    if image_.check_path is not None:
        wait_for_status_code(str(URL(external_url).with_path(image_.check_path)), 200)

    bridge_url = f'http://{get_bridge_ip(microservice)}:{image_.port}'
    logger.debug('Internal URL of docker image %s is %s', image_.image, bridge_url)

    return external_url, bridge_url


def _with_hea_env_vars(container_spec: DockerContainerSpec, registry_url: Optional[str], mongo_hostname: Optional[str],
                       mongo_hea_database: Optional[str],
                       mongo_hea_username: Optional[str], mongo_hea_password: Optional[str]) -> DockerContainerSpec:
    """
    Copies the provided container_spec, adding the environment variables corresponding to the provided arguments.

    :param container_spec: the image and configuration.
    :param registry_url: the URL of the registry service, which populates the HEASERVER_REGISTRY_URL environment
    variable.
    :param mongo_hostname: the hostname of the MongoDB service, which populates the MONGO_HOSTNAME environment variable.
    :param mongo_hea_database: the MongoDB database name, which populates the MONGO_HEA_DATABASE environment variable.
    :param mongo_hea_username: the MongoDB username, which populates the MONGO_HEA_USERNAME environment variable.
    :param mongo_hea_password: the MongoDB password, which populates the MONGO_HEA_PASSWORD environment variable.
    :return: the copy of the provided container_spec.
    """
    env_vars: Dict[str, str] = {}
    if mongo_hea_database is not None:
        env_vars['MONGO_HEA_DATABASE'] = mongo_hea_database
    if mongo_hea_username is not None:
        env_vars['MONGO_HEA_USERNAME'] = mongo_hea_username
    if mongo_hea_password is not None:
        env_vars['MONGO_HEA_PASSWORD'] = mongo_hea_password
    if registry_url is not None:
        env_vars['HEASERVER_REGISTRY_URL'] = registry_url
    if mongo_hostname is not None:
        env_vars['MONGO_HOSTNAME'] = mongo_hostname
    return container_spec.with_env_vars(env_vars)



"""
This module implements a simple API for launching a swagger UI for trying out a HEA microservice's REST APIs.
"""
import logging

from testcontainers.mongodb import MongoDbContainer
from contextlib import ExitStack
from . import runner, wstl, db
from .db import mongo
from aiohttp_swagger3 import SwaggerDocs, SwaggerUiSettings
from aiohttp_swagger3.handlers import application_json
from aiohttp import web
from importlib.metadata import version
from typing import Dict, List, Tuple, Callable, Iterable, Optional
from types import ModuleType
import os
from .docker import get_exposed_port, DockerImages
from heaserver.service.docker import DockerContainerSpec, start_microservice_container
from heaobject.root import HEAObjectDict
from copy import deepcopy

from .openapi import download_openapi_spec


def run(project_slug: str,
        fixtures: Dict[str, List[HEAObjectDict]],
        module: ModuleType,
        routes: Iterable[Tuple[Callable, str, Callable]],
        registry_docker_image: Optional[str] = None,
        other_docker_images: Optional[List[DockerContainerSpec]] = None) -> None:
    """
    Launches a swagger UI for trying out the given HEA APIs. It downloads the HEA OpenAPI spec from gitlab.com. It
    launches Docker containers for any other projects specified in the other_docker_images argument. It also
    launches two MongoDB databases in Docker containers, one for the project being tested, and the other for the other
    containers. It inserts the given HEA objects into both databases, creates HEA Server Registry entries for all
    containers, and makes the given routes available to query in swagger. All containers run on Docker's default bridge
    network.

    This function creates two MongoDB databases because the project being tested is not running in a Docker container,
    so it needs any other containers' external URLs (starting with localhost) rather than Docker's internal network's
    IP addresses. To simplify this function's arguments, we load the exact same data into both MongoDB databases,
    except for the registry components, which differ only in their base URLs and names. The latter is set to the
    component's base URL. One of the databases will contain localhost base URLs, and the other will contain Docker
    bridge network IP addresses. This function logs each MongoDB container's mongo connection string at the info level
    so that you can tell which container is which.

    :param project_slug: the Gitlab project slug of interest. Required.
    :param fixtures: a mapping of mongo collection -> list of HEA objects as dicts. Required.
    :param module: the microservice's service module.
    :param routes: a list of three-tuples containing the command, path and collable of each route of interest. The
    commands are one of: aiohttp.web.get, aiohttp.web.delete, aiohttp.web.post, aiohttp.web.put or aiohttp.web.view.
    :param registry_docker_image: an heaserver-registry docker image in REPOSITORY:TAG format, that will be launched
    after the MongoDB container is live.
    :param other_docker_images: optional list of ContainerSpec objects, which specify additional HEA microservice
    Docker containers to start.
    :raises OSError: If an error occurred accessing the OpenAPI spec.
    """
    if other_docker_images and registry_docker_image is None:
        raise ValueError('registry_docker_image must be non-None when other_docker_images is not empty')

    bridge_fixtures, external_fixtures = deepcopy(fixtures), deepcopy(fixtures)
    os.environ['MONGO_DB'] = 'hea'
    logger = logging.getLogger(__name__)
    with ExitStack() as stack, download_openapi_spec() as open_api_spec_file:
        external_mongo, external_mongodb_connection_string = _start_mongo(stack)
        logger.info('Mongo for the project being tested has connection string %s', external_mongodb_connection_string)

        if other_docker_images or registry_docker_image is not None:
            bridge_mongo, bridge_mongodb_connection_string = _start_mongo(stack)
        else:
            bridge_mongo = None
            bridge_mongodb_connection_string = None

        if registry_docker_image is not None:
            _, bridge_registry_url = start_microservice_container(
                DockerContainerSpec(image=registry_docker_image, port=8080, check_path='/components'), bridge_mongo,
                stack)
            external_registry_url, _ = start_microservice_container(
                DockerContainerSpec(image=registry_docker_image, port=8080, check_path='/components'), external_mongo,
                stack)
        else:
            external_registry_url = None
            bridge_registry_url = None

        if other_docker_images:
            logger.info('Bridge mongo for any other projects has connection string %s',
                        bridge_mongodb_connection_string)
            _start_other_docker_containers(bridge_fixtures, external_fixtures, bridge_mongo, other_docker_images,
                                           bridge_registry_url, stack)
            _insert_fixtures_into_db(bridge_mongo, bridge_fixtures)

        _start_project(external_fixtures, external_mongo, external_mongodb_connection_string, external_registry_url,
                       module, open_api_spec_file, project_slug, routes)


def _start_project(external_fixtures, external_mongo, external_mongodb_connection_string, external_registry_url, module,
                   open_api_spec_file, project_slug, routes):
    config_file = _generate_config_file(external_mongodb_connection_string, external_registry_url)
    config = runner.init(config_string=config_file)
    _insert_fixtures_into_db(external_mongo, external_fixtures)
    app = runner.get_application(db.mongo.Mongo,
                                 wstl_builder_factory=wstl.builder_factory(module.__package__, href='/'),
                                 config=config)
    _init_swagger_docs(app, open_api_spec_file, project_slug, routes)
    web.run_app(app)


def _start_other_docker_containers(bridge_fixtures: Dict[str, List[HEAObjectDict]],
                                   external_fixtures: Dict[str, List[HEAObjectDict]], mongo_: MongoDbContainer,
                                   other_docker_images: Optional[List[DockerContainerSpec]],
                                   registry_url: Optional[str], stack: ExitStack):
    for img in other_docker_images or []:
        external_url, bridge_url = start_microservice_container(img, mongo_, stack, registry_url)
        bridge_fixtures.setdefault('components', []).append(
            {'type': 'heaobject.registry.Component', 'base_url': bridge_url, 'name': bridge_url,
             "owner": "system|none",
             'resources': [r.to_dict() for r in img.resources or []]})
        external_fixtures.setdefault('components', []).append(
            {'type': 'heaobject.registry.Component', 'base_url': external_url, 'name': external_url,
             "owner": "system|none",
             'resources': [r.to_dict() for r in img.resources or []]})


def _init_swagger_docs(app: web.Application, openapi_spec_file: Optional[str], project_slug: str,
                       routes: Iterable[Tuple[Callable, str, Callable]]) -> None:
    swagger = SwaggerDocs(app,
                          swagger_ui_settings=SwaggerUiSettings(path="/docs"),
                          title=project_slug,
                          version=version(project_slug),
                          components=openapi_spec_file)
    swagger.register_media_type_handler('application/vnd.collection+json', application_json)
    swagger.add_routes([r[0](r[1], r[2]) for r in routes])


def _generate_config_file(mongodb_connection_string: str, registry_url: Optional[str]) -> str:
    if registry_url is None:
        config_file = f"""
[MongoDB]
ConnectionString = {mongodb_connection_string}
"""
    else:
        config_file = f"""
[DEFAULT]
Registry={registry_url}

[MongoDB]
ConnectionString = {mongodb_connection_string}
                """
    return config_file


def _start_mongo(stack: ExitStack) -> Tuple[MongoDbContainer, str]:
    mongo_container = MongoDbContainer(DockerImages.MONGODB.value)
    mongo_ = stack.enter_context(mongo_container)
    mongodb_connection_string = f'mongodb://test:test@{mongo_.get_container_host_ip()}:{get_exposed_port(mongo_, 27017)}/hea?authSource=admin'
    return mongo_, mongodb_connection_string


def _insert_fixtures_into_db(mongo_: MongoDbContainer, fixtures: Dict[str, List[HEAObjectDict]]) -> None:
    db_ = mongo_.get_connection_client().hea
    for k in fixtures or {}:
        db_[k].insert_many(mongo.replace_id_with_object_id(f) for f in fixtures[k])



"""
The representor package contains modules that transform data between WeSTL and a RESTful API output format, and a
factory for getting an object that outputs in the format with a specified mime type. Each representor module implements
a different output format and implements the following:

(required) MIME_TYPE: the format's mime type.

(optional) async def formats(wstl_obj, coll_url, dumps=json.dumps): formats a run-time WeSTL document into the output
format. This function may not be implemented for input formats that do not support data output.

(optional) async def parses(request): parses an HTTP request into a dict with NVP (name-value
 pair JSON. This function may not be implemented for output formats that do not support data input.

The representor concept comes from:
Amundsen, Mike. RESTful Web Clients. Sebastopol, CA: O'Reilly Media, Inc., 2017.

Use the package as follows:
>>> from heaserver.service.representor import factory
>>> factory.get_representor('application/json')
"""



"""
Utility classes and functions for working with aiohttp.
"""

from contextlib import AbstractContextManager
from multiprocessing.connection import Connection
from io import BytesIO
from aiohttp.streams import StreamReader
from typing import Protocol


class AioHTTPFileLikeObjectWrapper:
    """
    Wraps aiohttp.streams.StreamReader.read_nowait in a read() method so that it can be passed where file-like objects
    are expected.
    """
    def __init__(self, reader: StreamReader):
        """
        Creates a new AioHTTPFileLikeObjectManager object, passing in the stream reader to wrap.

        :param reader: a aiohttp.streams.StreamReader (required).
        """
        if reader is None:
            raise ValueError('reader cannot be None')
        self.__reader = reader

    def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        return self.__reader.read_nowait(n)


class ConnectionFileLikeObjectWrapper(AbstractContextManager):
    """
    Wraps a multiprocessing.connection.Connection object and provides file-like object methods.

    This class is a context manager, so it can be used in with statements.
    """
    def __init__(self, conn: Connection):
        """
        Creates a new ConnectionFileLikeObjectWrapper object, passing in the connection to wrap.

        :param conn: a multiprocessing.connection.Connection object (required).
        """
        if conn is None:
            raise ValueError('conn cannot be None')
        self.__conn = conn
        self.__buffer = BytesIO()

    def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        if len(b := self.__buffer.read(n)) > 0:
            return b
        try:
            result: bytes = self.__conn.recv_bytes()
            if len(result) > n:
                pos = self.__buffer.tell()
                self.__buffer.write(result[n:])
                self.__buffer.seek(pos)
                return result[:n]
            else:
                return result
        except EOFError:
            return b''

    def write(self, b):
        """
        Sends some bytes to the connection.

        :param b: some bytes (required).
        """
        self.__conn.send_bytes(b)

    def fileno(self):
        """
        Returns the integer file descriptor that is used by the connection.

        :return: the integer file descriptor.
        """
        return self.__conn.fileno()

    def close(self):
        """
        Closes the connection and any other resources associated with this object.
        """
        try:
            self.__buffer.close()
            self.__conn.close()
            self.__conn = None
        finally:
            if self.__conn is not None:
                try:
                    self.conn.close()
                except OSError:
                    pass

    def __exit__(self, *exc_details):
        self.close()


class SupportsAsyncRead(Protocol):
    """
    Protocol with an async read() method and a close() method.
    """
    async def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        pass

    def close(self):
        """
        Closes any resources associated with this object.
        """
        pass


class AsyncReader:
    """
    Wraps a bytes object in a simple reader with an asynchronous read method and a close method.
    """
    def __init__(self, b: bytes):
        """
        Creates a new AsyncReader, passing in a bytes object.

        :param b: bytes (required).
        """
        self.__b = BytesIO(b)

    async def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        return self.__b.read(n)

    def close(self):
        """
        Closes any resources associated with this object.
        """
        self.__b.close()


class StreamReaderWrapper:
    """
    Wraps an aiohttp StreamReader in an asyncio StreamReader-like object with a read() method and a no-op close()
    method.
    """
    def __init__(self, reader: StreamReader):
        if reader is None:
            raise ValueError('reader cannot be None')
        self.__reader = reader

    async def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        return await self.__reader.read(n)

    def close(self):
        pass



"""
Documentation for setup.py files is at https://setuptools.readthedocs.io/en/latest/setuptools.html
"""

from setuptools import setup, find_namespace_packages


# Import the README.md file contents
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='heaserver',
      version='1.0.0a57',
      description='The server side of HEA.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://risr.hci.utah.edu',
      author='Research Informatics Shared Resource, Huntsman Cancer Institute, Salt Lake City, UT',
      author_email='Andrew.Post@hci.utah.edu',
      python_requires='>=3.8',
      package_dir={'': 'src'},
      packages=find_namespace_packages(where='src'),
      package_data={'heaserver.service': ['py.typed']},
      install_requires=[
          'heaobject~=1.0.0a24',
          'aiohttp[speedups]~=3.8.1',
          'aiohttp-remotes~=1.2.0',
          'motor~=2.5.1',
          'pytz~=2021.3', # We should get rid of this dependency and switch to the built-in zoneinfo module after we no longer need python 3.8 support.
          'tzlocal~=4.1',
          'uritemplate~=4.1.1',
          'accept-types~=0.4.1',
          'mongoquery~=1.3.6',
          'jsonschema~=4.4.0',
          'jsonmerge~=1.8.0',
          'requests>=2.27',
          'types-requests>=2.27',  # Should be set at same version as requests.
          'boto3==1.20.39',
          'botocore==1.23.39'
      ],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Natural Language :: English',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: Implementation :: CPython',
          'Topic :: Software Development',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Scientific/Engineering :: Medical Science Apps.'
      ]
      )



"""
Sets up a testing environment for testing HEA services, including mock modules and classes. To use this module, do the
following in order:
1) Import this module before importing anything else to insure that all mocks are are used in subsequently imported
modules.
2) Import the service to be tested.
3) Call the get_test_case_cls with the fixtures to use in the unit tests.
4) Create subclasses of the returned unit test class to implement actual test cases. The mixin module in this package
contains implementations of unit tests to run for testing GET, POST, PUT and DELETE. Add these mixins as superclasses
of your test cases.
"""

from ..db.mockmongo import MockMongo
from . import mongotestcase
from aiohttp import web
import configparser
import logging
from typing import Type, Dict, List, Any, Optional, Union
from heaserver.service import runner, wstl
from .expectedvalues import expected_values, ActionSpec, LinkSpec
from yarl import URL
from heaobject.root import HEAObjectDict
from heaobject.user import NONE_USER


def get_test_case_cls(href: Union[str, URL],
                      wstl_package: str,
                      coll: str,
                      fixtures: Dict[str, List[HEAObjectDict]],
                      content: Optional[Dict[str, Dict[str, bytes]]] = None,
                      content_type: Optional[str] = None,
                      put_content_status: Optional[int] = None,
                      body_post: Optional[Dict[str, Dict[str, Any]]] = None,
                      body_put: Optional[Dict[str, Dict[str, Any]]] = None,
                      expected_one: Optional[Dict[str, Any]] = None,
                      expected_one_wstl: Optional[Dict[str, Any]] = None,
                      expected_one_duplicate_form: Optional[Dict[str, Any]] = None,
                      expected_all: Optional[Dict[str, Any]] = None,
                      expected_all_wstl: Optional[Dict[str, Any]] = None,
                      expected_opener: Optional[Union[str, URL]] = None,
                      expected_opener_body: Optional[Dict[str, Any]] = None,
                      sub: Optional[str] = NONE_USER) -> Type[mongotestcase.MongoTestCase]:
    """
    This function configures mocks. It returns a base test case class for testing a mongodb-based service with the
    provided fixtures. The test case class is a subclass of aiohttp.test_utils.AioHTTPTestCase, and the provided
    fixtures can be found in the following fields: _resource_path, _body_post, _body_put, _expected_all, and
    _expected_one.

    :param href: the resource being tested (required).
    :param wstl_package: the name of the package containing the wstl data package (required).
    :param coll: the MongoDB collection (required).
    :param fixtures: data to insert into the mongo database before running each test. Should be a dict of
    collection -> list of objects. Required.
    :param content: :param content: data to insert into GridFS (optional), as a collection string -> HEA Object id ->
    content as a bytes, bytearray, or array.array object.
    :param content_type: the MIME type of the content (optional).
    :param put_content_status: the expected status code for updating the content of the HEA object (optional). Normally
    should be 204 if the content is updatable and 405 if not. Default is None, which will cause associated tests to be
    skipped.
    :param body_post: JSON dict for data to be posted.
    :param body_put: JSON dict for data to be put. If None, the value of body_post will be used for PUTs.
    :param expected_one: The expected JSON dict for GET calls. If None, the value of expected_all will be used.
    :param expected_one_wstl: The expected JSON dict for GET calls that return the
    application/vnd.wstl+json mime type.
    :param expected_one_duplicate_form: The expected JSON dict for GET calls that return the
    object's duplicate form.
    :param expected_all: The expected JSON dict for GET-all calls.
    :param expected_all_wstl: The expected JSON dict for GET-all calls that return the
    application/vnd.wstl+json mime type.
    :param expected_opener: The expected URL of the resource that does the opening.
    :param expected_opener_body: The expected JSON dict for GET calls for an HEA desktop object's opener choices.
    """

    def mock_mongo(app: web.Application, config: configparser.ConfigParser):
        return MockMongo(app, config, fixtures, content)

    class MockMongoTestCase(mongotestcase.MongoTestCase):
        """
        Test case class for testing a mongodb-based service.
        """

        def __init__(self, methodName=None):
            """
            Initializes a test case.

            :param methodName: the name of the method to test.
            """
            super().__init__(methodName=methodName,
                             coll=coll,
                             href=href,
                             body_post=body_post,
                             body_put=body_put,
                             expected_all=expected_all,
                             expected_one=expected_one or expected_all,
                             expected_one_wstl=expected_one_wstl or expected_all_wstl,
                             expected_all_wstl=expected_all_wstl,
                             expected_one_duplicate_form=expected_one_duplicate_form,
                             expected_opener=expected_opener,
                             expected_opener_body=expected_opener_body,
                             wstl_package=wstl_package,
                             content=content,
                             content_type=content_type,
                             put_content_status=put_content_status,
                             sub=sub)

        def run(self, result=None):
            with self._caplog.at_level(logging.DEBUG):
                return super().run(result)

        async def get_application(self):
            return runner.get_application(db=mock_mongo,
                                          wstl_builder_factory=wstl.builder_factory(wstl_package, href=href),
                                          testing=True)

    return MockMongoTestCase


def get_test_case_cls_default(coll: str,
                              wstl_package: str,
                              fixtures: Dict[str, List[HEAObjectDict]],
                              duplicate_action_name: str,
                              content: Optional[Dict[str, Dict[str, bytes]]] = None,
                              content_type: Optional[str] = None,
                              put_content_status: Optional[int] = None,
                              include_root=False,
                              href: Optional[Union[str, URL]] = None,
                              get_actions: Optional[List[ActionSpec]] = None,
                              get_all_actions: Optional[List[ActionSpec]] = None,
                              expected_opener: Optional[LinkSpec] = None,
                              sub: Optional[str] = NONE_USER) -> Type[mongotestcase.MongoTestCase]:
    if href is None:
        href_ = str(URL(f'/{coll}/'))
    else:
        href_ = str(href)
        if not href_.endswith('/'):
            href_ = href_ + '/'
    return get_test_case_cls(href=href_, wstl_package=wstl_package, coll=coll, fixtures=fixtures,
                             content=content, content_type=content_type, put_content_status=put_content_status, sub=sub,
                             **expected_values(fixtures, coll, wstl.builder(package=wstl_package),
                                               duplicate_action_name, href_,
                                               include_root=include_root, get_actions=get_actions,
                                               get_all_actions=get_all_actions,
                                               opener_link=expected_opener))
